# Copyright 2014 OpenStack Foundation
# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import queue
import threading


class MergedQueue(object):
    def __init__(self):
        self.queue = collections.deque()
        self.lock = threading.RLock()
        self.condition = threading.Condition(self.lock)
        self.join_condition = threading.Condition(self.lock)
        self.tasks = 0

    def qsize(self):
        return len(self.queue)

    def empty(self):
        return self.qsize() == 0

    def put(self, item):
        # Returns the original item if added, or an updated equivalent
        # item if already enqueued.
        self.condition.acquire()
        ret = None
        try:
            for x in self.queue:
                if item == x:
                    ret = x
                    if hasattr(ret, 'merge'):
                        ret.merge(item)
            if ret is None:
                ret = item
                self.queue.append(item)
                self.condition.notify()
        finally:
            self.condition.release()
        return ret

    def get(self):
        self.condition.acquire()
        try:
            while True:
                try:
                    ret = self.queue.popleft()
                    self.join_condition.acquire()
                    self.tasks += 1
                    self.join_condition.release()
                    return ret
                except IndexError:
                    self.condition.wait()
        finally:
            self.condition.release()

    def task_done(self):
        self.join_condition.acquire()
        self.tasks -= 1
        self.join_condition.notify()
        self.join_condition.release()

    def join(self):
        self.join_condition.acquire()
        while self.tasks:
            self.join_condition.wait()
        self.join_condition.release()


class NamedQueue(queue.Queue):
    """For identification of queues in logs"""

    def __init__(self, name, maxsize=0):
        super().__init__(maxsize)
        self.name = name

    def __repr__(self):
        return f"<Queue {self.name} [{id(self)}]>"



# Copyright 2015 Puppet Labs
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import voluptuous as v
import time

from zuul import model
from zuul.lib.logutil import get_annotated_logger
from zuul.reporter import BaseReporter
from zuul.exceptions import MergeFailure
from zuul.driver.util import scalar_or_list
from zuul.driver.github.githubsource import GithubSource


class GithubReporter(BaseReporter):
    """Sends off reports to Github."""

    name = 'github'
    log = logging.getLogger("zuul.GithubReporter")

    # Merge modes supported by github
    merge_modes = {
        model.MERGER_MERGE: 'merge',
        model.MERGER_MERGE_RESOLVE: 'merge',
        model.MERGER_MERGE_RECURSIVE: 'merge',
        model.MERGER_MERGE_ORT: 'merge',
        model.MERGER_SQUASH_MERGE: 'squash',
        model.MERGER_REBASE: 'rebase',
    }

    def __init__(self, driver, connection, pipeline, config=None,
                 parse_context=None):
        super(GithubReporter, self).__init__(
            driver, connection, config, parse_context)
        self._commit_status = self.config.get('status', None)
        self._create_comment = self.config.get('comment', True)
        self._check = self.config.get('check', False)
        self._merge = self.config.get('merge', False)
        self._labels = self.config.get('label', [])
        if not isinstance(self._labels, list):
            self._labels = [self._labels]
        self._unlabels = self.config.get('unlabel', [])
        self._review = self.config.get('review')
        self._review_body = self.config.get('review-body')
        if not isinstance(self._unlabels, list):
            self._unlabels = [self._unlabels]

    def getContext(self, manager):
        return "{}/{}".format(manager.tenant.name,
                              manager.pipeline.name)

    def report(self, item, phase1=True, phase2=True):
        """Report on an event."""
        log = get_annotated_logger(self.log, item.event)

        ret = []
        for change in item.changes:
            err = self._reportChange(item, change, log, phase1, phase2)
            if err:
                ret.append(err)
        return ret

    def _reportChange(self, item, change, log, phase1=True, phase2=True):
        """Report on an event."""
        # If the source is not GithubSource we cannot report anything here.
        if not isinstance(change.project.source, GithubSource):
            return

        # For supporting several Github connections we also must filter by
        # the canonical hostname.
        if change.project.source.connection.canonical_hostname != \
                self.connection.canonical_hostname:
            return

        # order is important for github branch protection.
        # A status should be set before a merge attempt
        if phase1 and self._commit_status is not None:
            if (hasattr(change, 'patchset') and
                    change.patchset is not None):
                self.setCommitStatus(item, change)
            elif (hasattr(change, 'newrev') and
                    change.newrev is not None):
                self.setCommitStatus(item, change)
        # Comments, labels, and merges can only be performed on pull requests.
        # If the change is not a pull request (e.g. a push) skip them.
        if hasattr(change, 'number'):
            errors_received = False
            if phase1:
                if self._labels or self._unlabels:
                    self.setLabels(item, change)
                if self._review:
                    self.addReview(item, change)
                if self._check:
                    check_errors = self.updateCheck(item, change)
                    # TODO (felix): We could use this mechanism to
                    # also report back errors from label and review
                    # actions
                    if check_errors:
                        item.current_build_set.warning_messages.extend(
                            check_errors
                        )
                        errors_received = True
                if self._create_comment or errors_received:
                    self.addPullComment(item, change)
            if phase2 and self._merge:
                try:
                    self.mergePull(item, change)
                except Exception as e:
                    self.addPullComment(item, change, str(e))

    def _formatJobResult(self, job_fields):
        # We select different emojis to represents build results:
        # heavy_check_mark: SUCCESS
        # warning: SKIPPED/ABORTED
        # x: all types of FAILUREs
        # In addition, failure results are in bold text

        job_result = job_fields[2]
        # Also need to handle user defined success_message.
        # The job_fields[6]: the user defined seccess_message (if available)
        success_message = job_fields[6]

        emoji = 'x'
        bold_result = True

        if job_result in ('SUCCESS', success_message):
            emoji = 'heavy_check_mark'
            bold_result = False
        elif job_result in ('SKIPPED', 'ABORTED', 'CANCELED'):
            emoji = 'warning'
            bold_result = False

        if bold_result:
            return ':%s: [%s](%s) **%s**%s%s%s\n' % (
                (emoji,) + job_fields[:6])
        else:
            return ':%s: [%s](%s) %s%s%s%s\n' % (
                (emoji,) + job_fields[:6])

    def _formatItemReportJobs(self, item, short):
        # Return the list of jobs portion of the report
        ret = ''
        jobs_fields, skipped = self._getItemReportJobsFields(item)
        for job_fields in jobs_fields:
            ret += self._formatJobResult(job_fields)
        if skipped:
            jobtext = 'job' if skipped == 1 else 'jobs'
            ret += 'Skipped %i %s\n' % (skipped, jobtext)
        return ret

    def addPullComment(self, item, change, comment=None):
        log = get_annotated_logger(self.log, item.event)
        message = comment or self._formatItemReport(item)
        project = change.project.name
        pr_number = change.number
        log.debug('Reporting change %s, params %s, message: %s',
                  change, self.config, message)
        self.connection.commentPull(project, pr_number, message,
                                    zuul_event_id=item.event)

    def setCommitStatus(self, item, change):
        log = get_annotated_logger(self.log, item.event)

        project = change.project.name
        if hasattr(change, 'patchset'):
            sha = change.patchset
        elif hasattr(change, 'newrev'):
            sha = change.newrev
        state = self._commit_status

        url = item.formatItemUrl()

        description = '%s status: %s' % (item.manager.pipeline.name,
                                         self._commit_status)

        if len(description) >= 140:
            # This pipeline is named with a long name and thus this
            # desciption would overflow the GitHub limit of 1024 bytes.
            # Truncate the description. In practice, anything over 140
            # characters seems to trip the limit.
            description = 'status: %s' % self._commit_status

        log.debug(
            'Reporting change %s, params %s, '
            'context: %s, state: %s, description: %s, url: %s',
            change, self.config, self.getContext(item.manager),
            state, description, url)

        self.connection.setCommitStatus(
            project, sha, state, url, description,
            self.getContext(item.manager),
            zuul_event_id=item.event)

    def mergePull(self, item, change):
        log = get_annotated_logger(self.log, item.event)
        merge_mode = item.current_build_set.getMergeMode(change)

        if merge_mode not in self.merge_modes:
            mode = model.get_merge_mode_name(merge_mode)
            self.log.warning('Merge mode %s not supported by Github', mode)
            raise MergeFailure('Merge mode %s not supported by Github' % mode)

        project = change.project.name
        pr_number = change.number
        sha = change.patchset
        log.debug('Reporting change %s, params %s, merging via API',
                  change, self.config)
        message = self._formatMergeMessage(change, merge_mode)
        merge_mode = self.merge_modes[merge_mode]

        max_retries = 5
        for i in range(1, max_retries + 1):
            try:
                merge_sha = self.connection.mergePull(
                    project, pr_number, message, sha=sha,
                    method=merge_mode, zuul_event_id=item.event)
                self.connection.updateChangeAttributes(
                    change,
                    merge_commit_sha=merge_sha,
                    is_merged=True)
                return
            except MergeFailure as e:
                log.exception('Merge attempt of change %s  %s/%s failed.',
                              change, i, max_retries, exc_info=True)
                error_message = str(e)
                if i < max_retries:
                    time.sleep(5)
                    # Sometimes GitHub lies about the merge status and reports
                    # merge failed but in fact did the merge successfully. To
                    # account for that re-fetch the PR and check if it's still
                    # unmerged.
                    try:
                        if self.connection.isMerged(change, item.event):
                            # The change has been merged despite we got an
                            # error back from GitHub
                            log.warning(
                                'Merge attempt of change %s failed but change '
                                'is merged. Treating as success.', change)
                            self.connection.updateChangeAttributes(
                                change, is_merged=True)
                            return
                    except Exception:
                        log.exception('Error while checking for merged '
                                      'change %s', change)
        log.warning('Merge of change %s failed after %s attempts, giving up',
                    change, max_retries)
        raise MergeFailure(error_message)

    def addReview(self, item, change):
        log = get_annotated_logger(self.log, item.event)
        project = change.project.name
        pr_number = change.number
        sha = change.patchset
        log.debug('Reporting change %s, params %s, review:\n%s',
                  change, self.config, self._review)
        self.connection.reviewPull(
            project,
            pr_number,
            sha,
            self._review,
            self._review_body,
            zuul_event_id=item.event)
        for label in self._unlabels:
            self.connection.unlabelPull(project, pr_number, label,
                                        zuul_event_id=item.event)

    def updateCheck(self, item, change):
        log = get_annotated_logger(self.log, item.event)
        message = self._formatItemReport(item)
        project = change.project.name
        pr_number = change.number
        sha = change.patchset

        status = self._check
        # We declare a item as completed if it either has a result
        # (success|failure) or a dequeue reporter is called (cancelled in case
        # of Github checks API). For the latter one, the item might or might
        # not have a result, but we still must set a conclusion on the check
        # run. Thus, we cannot rely on the buildset's result only, but also
        # check the state the reporter is going to report.
        completed = (
            item.current_build_set.result is not None or status == "cancelled"
            or status == "skipped" or status == "neutral"
        )

        log.debug(
            "Updating check for change %s, params %s, context %s, message: %s",
            change, self.config, self.getContext(item.manager), message
        )

        details_url = item.formatItemUrl()

        # Check for inline comments that can be reported via checks API
        file_comments = self.getFileComments(item, change)

        # Github allows an external id to be added to a check run. We can use
        # this to identify the check run in any custom actions we define.
        # To uniquely identify the corresponding buildset in zuul, we need
        # tenant, pipeline and change. The buildset's uuid cannot be used
        # safely, as it might change e.g. during a gate reset. Fore more
        # information, please see Jim's comment on
        # https://review.opendev.org/#/c/666258/7
        external_id = json.dumps(
            {
                "tenant": item.manager.tenant.name,
                "pipeline": item.manager.pipeline.name,
                "change": change.number,
            }
        )

        state = item.dynamic_state[self.connection.connection_name]
        check_run_ids = state.setdefault('check_run_ids', {})
        check_run_id = check_run_ids.get(change.cache_key)
        check_run_id, errors = self.connection.updateCheck(
            project,
            pr_number,
            sha,
            status,
            completed,
            self.getContext(item.manager),
            details_url,
            message,
            file_comments,
            external_id,
            zuul_event_id=item.event,
            check_run_id=check_run_id,
        )

        if check_run_id:
            check_run_ids[change.cache_key] = check_run_id

        return errors

    def setLabels(self, item, change):
        log = get_annotated_logger(self.log, item.event)
        project = change.project.name
        pr_number = change.number
        if self._labels:
            log.debug('Reporting change %s, params %s, labels:\n%s',
                      change, self.config, self._labels)
        for label in self._labels:
            self.connection.labelPull(project, pr_number, label,
                                      zuul_event_id=item.event)
        if self._unlabels:
            log.debug('Reporting change %s, params %s, unlabels:\n%s',
                      change, self.config, self._unlabels)
        for label in self._unlabels:
            self.connection.unlabelPull(project, pr_number, label,
                                        zuul_event_id=item.event)

    def _formatMergeMessage(self, change, merge_mode):
        message = []
        # For squash merges we don't need to add the title to the body
        # as it will already be set as the commit subject.
        if merge_mode != model.MERGER_SQUASH_MERGE:
            if change.title:
                message.append(change.title)
        if change.body_text:
            message.append(change.body_text)
        merge_message = "\n\n".join(message)

        if change.reviews:
            review_users = []
            for r in change.reviews:
                name = r['by']['name']
                email = r['by']['email']
                username = r['by']['username']
                review_message = 'Reviewed-by:'
                if name:
                    review_message += ' {}'.format(name)
                elif username:
                    review_message += ' {}'.format(username)
                else:
                    review_message += ' Anonymous'
                if email:
                    review_message += ' <{}>'.format(email)
                review_users.append(review_message)
            merge_message += '\n\n'
            merge_message += '\n'.join(review_users)

        return merge_message

    def getSubmitAllowNeeds(self, manager):
        """Get a list of code review labels that are allowed to be
        "needed" in the submit records for a change, with respect
        to this queue.  In other words, the list of review labels
        this reporter itself is likely to set before submitting.
        """

        # check if we report a status or a check, if not we can return an
        # empty list
        status = self.config.get('status')
        check = self.config.get("check")
        if not any([status, check]):
            return []

        # we return a status so return the status we report to github
        return [self.getContext(manager)]


def getSchema():
    github_reporter = v.Schema({
        'status': v.Any('pending', 'success', 'failure'),
        'comment': bool,
        'merge': bool,
        'label': scalar_or_list(str),
        'unlabel': scalar_or_list(str),
        'review': v.Any('approve', 'request-changes', 'comment'),
        'review-body': str,
        'check': v.Any(
            "queued", "in_progress", "success", "failure", "cancelled",
            "skipped", "neutral"),
    })
    return github_reporter



# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""build_idx

Revision ID: 151893067f91
Revises: 0ed5def089e2
Create Date: 2023-07-05 11:50:03.931815

"""

# revision identifiers, used by Alembic.
revision = '151893067f91'
down_revision = '0ed5def089e2'
branch_labels = None
depends_on = None

from alembic import op


def upgrade(table_prefix=''):
    # Create indexes like:
    # artifact_build_idx on zuul_artifact.build_id
    # This happens automatically on foreign keys in mysql but must be
    # done explicitly in postgres.
    for suffix in ['artifact',
                   'provides',
                   'build_event']:
        op.create_index(
            f'{table_prefix}{suffix}_build_id_idx',
            f'{table_prefix}zuul_{suffix}', ['build_id'])
    op.create_index(
        f'{table_prefix}build_buildset_id_idx',
        f'{table_prefix}zuul_build', ['buildset_id'])


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2021-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import math

import zuul.provider.schema as provider_schema
from zuul.lib.voluputil import (
    AsList,
    Constant,
    Exclusive,
    Nullable,
    Optional,
    Required,
    RequiredExclusive,
    assemble,
    discriminate,
)

import voluptuous as vs

from zuul.driver.azure.azureendpoint import (
    AzureCreateStateMachine,
    AzureDeleteStateMachine,
)
from zuul.model import QuotaInformation
from zuul.provider import (
    BaseProvider,
    BaseProviderFlavor,
    BaseProviderImage,
    BaseProviderLabel,
    BaseProviderSchema,
)

URL_VM_SIZES = 'https://azure.microsoft.com/en-us/global-infrastructure/services/?products=virtual-machines'  # noqa


class AzureProviderImage(BaseProviderImage):
    azure_image_reference = {
        Required(
            'sku',
            doc="""
            The image SKU.
            """,
        ): str,
        Required(
            'publisher',
            doc="""\
            The image Publisher.
            """,
        ): str,
        Required(
            'version',
            doc="""\
            The image version.
            """,
        ): str,
        Required(
            'offer',
            doc="""\
            The image offer.
            """,
        ): str,
    }
    azure_image_filter = {
        Optional(
            'location',
            doc="""\
            The image location.
            """,
        ): Nullable(str),
        Optional(
            'name',
            doc="""\
            The image name.
            """,
        ): Nullable(str),
        Optional(
            'tags',
            doc="""\
            The image tags.
            """,
        ): Nullable({str: str}),
    }
    azure_gallery_image = {
        Required(
            'gallery-name',
            doc="""\
            The image gallery name.
            """,
        ): str,
        Required(
            'name',
            doc="""\
            The image name.
            """,
        ): str,
        Optional(
            'version',
            doc="""\
            The image version.  Omit to use the latest version.
            """,
        ): Nullable(str),
    }
    # This is used here and in flavors and labels
    inheritable_azure_image_schema = assemble(
        vs.Schema({
            Optional(
                'volume-size',
                doc="""\
                The size of the operating system disk, in GiB.
                """,
            ): Nullable(int),
            Optional(
                'ephemeral-disk',
                doc="""\
                If set to ``true``, Azure will create an ephemeral OS disk
                instead of a managed disk.
                """,
            ): Nullable(bool),
            Optional(
                'generate-password',
                doc="""\
                If booting a Windows image, an administrative password is
                required.  If the password is not actually used (e.g., the
                image has key-based authentication enabled), a random
                password can be provided by enabling this option.  The
                password is not stored anywhere and is not retrievable.
                """,
            ): Nullable(bool),
        }),
        provider_schema.cloud_image,
    )
    azure_cloud_schema = vs.Schema({
        Exclusive(Required(
            'image-filter',
            doc="""\
            Specifies a private image to use via filters.  Either this field,
            :attr:`provider[azure].images[cloud].shared-gallery-image`,
            :attr:`provider[azure].images[cloud].community-gallery-image`,
            :attr:`provider[azure].images[cloud].image-reference`, or
            :attr:`provider[azure].images[cloud].image-id` must be
            provided.

            If a filter is provided, Zuul will list all of the images
            in the provider's resource group and reduce the list using
            the supplied filter.  All items specified in the filter must
            match in order for an image to match.  If more than one image
            matches, the images are sorted by name and the last one
            matches.

            The following filters are available:
            """,
        ), 'spec'): azure_image_filter,
        Exclusive(Required(
            'image-id',
            doc="""\
            Specifies a private image to use by ID.  Either this field,
            :attr:`provider[azure].images[cloud].shared-gallery-image`,
            :attr:`provider[azure].images[cloud].community-gallery-image`,
            :attr:`provider[azure].images[cloud].image-reference`, or
            :attr:`provider[azure].images[cloud].image-filter` must be
            provided.
            """,
        ), 'spec'): str,
        Exclusive(Required(
            'image-reference',
            doc="""\
            Specifies a public image to use by reference.  Either this field,
            :attr:`provider[azure].images[cloud].shared-gallery-image`,
            :attr:`provider[azure].images[cloud].community-gallery-image`,
            :attr:`provider[azure].images[cloud].image-id`, or
            :attr:`provider[azure].images[cloud].image-filter` must be
            provided.
            """,
        ), 'spec'): azure_image_reference,
        Exclusive(Required(
            'community-gallery-image',
            doc="""\
            Specifies a community gallery image to use.  Either this field,
            :attr:`provider[azure].images[cloud].shared-gallery-image`,
            :attr:`provider[azure].images[cloud].image-reference`,
            :attr:`provider[azure].images[cloud].image-id`, or
            :attr:`provider[azure].images[cloud].image-filter` must be
            provided.
            """,
        ), 'spec'): azure_gallery_image,
        Exclusive(Required(
            'shared-gallery-image',
            doc="""\
            Specifies a shared gallery image to use.  Either this field,
            :attr:`provider[azure].images[cloud].community-gallery-image`,
            :attr:`provider[azure].images[cloud].image-reference`,
            :attr:`provider[azure].images[cloud].image-id`, or
            :attr:`provider[azure].images[cloud].image-filter` must be
            provided.
            """,
        ), 'spec'): azure_gallery_image,
    })
    cloud_schema = vs.All(
        assemble(
            BaseProviderImage.cloud_schema,
            azure_cloud_schema,
            inheritable_azure_image_schema,
        ),
        RequiredExclusive(
            'image_id', 'image_reference', 'image_filter',
            'community_gallery_image', 'shared_gallery_image',
            msg=('Provide one of "image-id", '
                 '"image-reference", "image-filter", '
                 '"community-gallery-image", or '
                 '"shared-gallery-image" keys'))
    )
    zuul_schema = assemble(
        BaseProviderImage.zuul_schema,
        inheritable_azure_image_schema,
    )
    inheritable_cloud_schema = assemble(
        BaseProviderImage.inheritable_cloud_schema,
        inheritable_azure_image_schema,
    )
    inheritable_zuul_schema = assemble(
        BaseProviderImage.inheritable_zuul_schema,
        inheritable_azure_image_schema,
    )
    schema = vs.Union(
        cloud_schema, zuul_schema,
        discriminant=discriminate(
            lambda val, alt: val['type'] == alt['type'].validators[0])
    )
    inheritable_schema = assemble(
        inheritable_cloud_schema,
        inheritable_zuul_schema,
    )

    def __init__(self, image_config, provider_config):
        self.image_id = None
        self.image_reference = None
        self.image_filter = None
        self.community_gallery_image = None
        self.shared_gallery_image = None
        super().__init__(image_config, provider_config)
        self.format = 'vhd'
        # Implement provider defaults
        if self.connection_type is None:
            self.connection_type = 'ssh'
        if self.connection_port is None:
            self.connection_port = 22


class AzureProviderFlavor(BaseProviderFlavor):
    azure_flavor_schema = vs.Schema({
        Required(
            'vm-size',
            doc=f"""\
            Size of the VM to use in Azure.  See the `List of VM
            sizes`_ for the list of sizes availabile in each region.

            .. _`List of VM sizes`: {URL_VM_SIZES}
            """,
        ): str,
        # TODO: add "low" priority
        Optional(
            'priority',
            doc="""\
            Whether to create regular or spot instances.
            """,
            default='regular'
        ): vs.Any(
            Constant(
                'regular',
                doc="""\
                A regular instance.
                """,
            ),
            Constant(
                'spot',
                doc="""\
                A spot instance.
                """,
            ),
        ),
        Optional(
            'ipv4',
            doc="""\
            Whether to enable IPv4 networking.  Defaults to true unless IPv6
            is enabled.  Enabling this will attach a private IPv4 address.
            """,
            default=False): bool,
        Optional(
            'ipv6',
            doc="""\
            Whether to enable IPv6 networking.
            Enabling this will attach a private IPv6 address.
            """,
            default=False): bool,
    })

    inheritable_schema = assemble(
        BaseProviderFlavor.inheritable_schema,
        # This is already included via the image, but listed again
        # here for clarity.
        AzureProviderImage.inheritable_azure_image_schema,
        provider_schema.cloud_flavor,
    )
    schema = assemble(
        BaseProviderFlavor.schema,
        provider_schema.cloud_flavor,
        AzureProviderImage.inheritable_azure_image_schema,
        azure_flavor_schema,
    )


class AzureProviderLabel(BaseProviderLabel):
    azure_network_reference = {
        Optional(
            'resource-group',
            doc="""\
            The resource group that contains the subnet.
            """,
        ): Nullable(str),
        Required(
            'network',
            doc="""\
            The name of the subnet's network.
            """,
        ): str,
        Optional(
            'subnet',
            doc="""\
            The name of the subnet.
            """,
            default='default'): str,
    }

    azure_identity_reference = {
        Optional(
            'resource-group',
            doc="""\
            The resource group that contains the identity.
            """,
        ): Nullable(str),
        Required(
            'name',
            doc="""\
            The name of the identity.
            """,
        ): str,
    }

    azure_label_schema = vs.Schema({
        Exclusive(Required(
            'subnet-id',
            doc="""\
            Specifies the subnet to use by ID.
            """,
        ), 'subnet'): str,
        Exclusive(Required(
            'subnet-reference',
            doc="""\
            Specifies the subnet to use by reference
            """,
        ), 'subnet'): azure_network_reference,
        Optional(
            'user-assigned-identities',
            doc="""\
            """,
            default=[],
        ): AsList(azure_identity_reference),
        Optional(
            'key-data',
            doc="""\
            The SSH public key that should be installed on the node.
            """,
        ): Nullable(str),
    })

    inheritable_schema = assemble(
        BaseProviderLabel.inheritable_schema,
        # This is already included via the image, but listed again
        # here for clarity.
        AzureProviderImage.inheritable_azure_image_schema,
        provider_schema.host_key_checking,
        azure_label_schema,
    )

    schema = vs.All(
        assemble(
            BaseProviderLabel.schema,
            AzureProviderImage.inheritable_azure_image_schema,
            provider_schema.host_key_checking,
            azure_label_schema,
        ),
        RequiredExclusive(
            'subnet_id', 'subnet_reference',
            msg=('Provide one of "subnet-id" or '
                 '"subnet-reference" keys'))
    )

    image_flavor_inheritable_schema = assemble(
        AzureProviderImage.inheritable_azure_image_schema,
    )


class AzureProviderSchema(BaseProviderSchema):
    def getLabelSchema(self):
        return AzureProviderLabel.schema

    def getImageSchema(self):
        return AzureProviderImage.schema

    def getFlavorSchema(self):
        return AzureProviderFlavor.schema

    def getInheritableLabelSchema(self):
        return AzureProviderLabel.inheritable_schema

    def getInheritableImageSchema(self):
        return AzureProviderImage.inheritable_schema

    def getInheritableZuulImageSchema(self):
        return AzureProviderImage.inheritable_zuul_schema

    def getInheritableCloudImageSchema(self):
        return AzureProviderImage.inheritable_cloud_schema

    def getInheritableFlavorSchema(self):
        return AzureProviderFlavor.inheritable_schema

    def getProviderSchema(self):
        schema = super().getProviderSchema()

        resource_limits = {
            Optional(
                'instances',
                default=vs.UNDEFINED,
                doc="""The number of instances.""",
            ): int,
            Optional(
                'cores',
                default=vs.UNDEFINED,
                doc="""The number of cores used by regular instances.""",
            ): int,
            Optional(
                'ram',
                default=vs.UNDEFINED,
                doc="""The amount of ram, in MiB.""",
            ): int,
            Optional(
                'lowPriorityCores',
                default=vs.UNDEFINED,
                doc="""\
                The number of low priority cores (including spot instances).
                """,
            ): int,
        }

        azure_provider_schema = vs.Schema({
            Required(
                'region',
                doc="""Name of the Azure region to use.""",
            ): str,
            Required(
                'resource-group',
                doc="""\
                Name of the resource group in which to place nodes.
                """,
            ): str,
            Optional(
                'resource-limits',
                doc="""\
                Resource limits for this provider.  Configure these
                values to cause Zuul to attempt to limit the resource
                usage.  This can be used to limit Zuul's usage to a
                level below the cloud quota.
                """,
                default=dict()): resource_limits,
        })

        return assemble(
            schema,
            azure_provider_schema,
            doc="""\
            The attributes available for configuring an Azure provider
            are below.
            """,
        )


class AzureProvider(BaseProvider, subclass_id='azure'):
    log = logging.getLogger("zuul.AzureProvider")
    schema = AzureProviderSchema().getProviderSchema()

    @property
    def endpoint(self):
        ep = getattr(self, '_endpoint', None)
        if ep:
            return ep
        self._set(_endpoint=self.getEndpoint())
        return self._endpoint

    def parseImage(self, image_config, provider_config, connection):
        return AzureProviderImage(image_config, provider_config)

    def parseFlavor(self, flavor_config, provider_config, connection):
        return AzureProviderFlavor(flavor_config, provider_config)

    def parseLabel(self, label_config, provider_config, connection):
        return AzureProviderLabel(label_config, provider_config)

    def getEndpoint(self):
        return self.driver.getEndpoint(self)

    def getCreateStateMachine(self, node, image_external_id, log):
        # TODO: decide on a method of producing a hostname
        # that is max 15 chars.
        hostname = f"np{node.uuid[:13]}"
        label = self.labels[node.label]
        flavor = self.flavors[label.flavor]
        image = self.images[label.image]
        return AzureCreateStateMachine(
            self,
            self.endpoint,
            node,
            hostname,
            label,
            flavor,
            image,
            image_external_id,
            node.tags,
            log)

    def getDeleteStateMachine(self, node, log):
        return AzureDeleteStateMachine(self.endpoint, node, log)

    def getEndpointLimits(self):
        limits = self.endpoint.quota_cache.getLimits()
        if limits is None:
            limits = {}
        else:
            limits = limits.quota
        return QuotaInformation(default=math.inf, **limits)

    def getQuotaForLabel(self, label):
        flavor = self.flavors[label.flavor]
        return self.endpoint.getQuotaForLabel(label, flavor)

    def refreshQuotaForLabel(self, label, update):
        flavor = self.flavors[label.flavor]
        return self.endpoint.refreshQuotaForLabel(label, flavor, update)

    def getImageUploadJob(self, provider_image, image_name,
                          image_format, metadata, md5, sha256):
        return self.endpoint.getImageUploadJob(
            self.resource_group, provider_image, image_name,
            image_format, metadata, md5, sha256)

    def deleteImage(self, external_id):
        self.endpoint.deleteImage(external_id)



# Copyright 2014 Rackspace Australia
# Copyright 2021 BMW Group
# Copyright 2021, 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from enum import IntFlag
import logging
import json

from zuul.zk.zkobject import ZKContext, ShardedZKObject
from zuul.zk.locks import (
    SessionAwareReadLock,
    SessionAwareWriteLock,
    locked as zk_locked
)

from kazoo.exceptions import NoNodeError

# Default marker to raise an exception on cache miss in getProjectBranches()
RAISE_EXCEPTION = object()


# These flags should be the purview of the drivers, but we need to
# know about them in order to support backwards compatability to
# MODEL_API < 27.  In the future, we should be able to make these
# driver-specific and have driver-specific subclasses of BranchInfo,
# etc.
class BranchFlag(IntFlag):
    CLEAR = 0
    PRESENT = 0x1
    PROTECTED = 0x2
    LOCKED = 0x4


# A helper method for the branch cache below.
def return_default(default, project_name):
    if default is RAISE_EXCEPTION:
        raise LookupError(
            f"No branches for project {project_name}")
    return default


class BranchInfo:
    def __init__(self, name, present=None, protected=None, locked=None):
        self.name = name
        # These are tri-state: None means indeterminate, true or false
        # are definitive.
        self.present = present
        self.protected = protected
        self.locked = locked

    def update(self, other):
        if other.present is not None:
            self.present = other.present
        if other.protected is not None:
            self.protected = other.protected
        if other.locked is not None:
            self.locked = other.locked

    def toDict(self):
        # This doesn't really return a dict, but like other toDict
        # methods, it returns the object that will be encoded into
        # JSON.  It just happens we don't need a full dict for this.
        return [self.flags, self.valid_flags]

    @property
    def flags(self):
        flags = BranchFlag.CLEAR
        if self.present:
            flags |= BranchFlag.PRESENT
        if self.protected:
            flags |= BranchFlag.PROTECTED
        if self.locked:
            flags |= BranchFlag.LOCKED
        return flags

    @property
    def valid_flags(self):
        # If a flag is None, then we don't know it for this branch so
        # we consider it invalid.
        valid_flags = BranchFlag.CLEAR
        if self.present is not None:
            valid_flags |= BranchFlag.PRESENT
        if self.protected is not None:
            valid_flags |= BranchFlag.PROTECTED
        if self.locked is not None:
            valid_flags |= BranchFlag.LOCKED
        return valid_flags

    @classmethod
    def fromDict(cls, name, data):
        o = cls(name)
        flags = BranchFlag(data[0])
        valid_flags = BranchFlag(data[1])

        if BranchFlag.PRESENT in valid_flags:
            o.present = bool(flags & BranchFlag.PRESENT)
        if BranchFlag.PROTECTED in valid_flags:
            o.protected = bool(flags & BranchFlag.PROTECTED)
        if BranchFlag.LOCKED in valid_flags:
            o.locked = bool(flags & BranchFlag.LOCKED)
        return o


class ProjectInfo:
    """Store branch cache project information in ZK

    If a project is absent from the cache, it needs to be queried from
    the source.
    """
    def __init__(self, name, merge_modes=None, default_branch=None):
        self.name = name
        self.merge_modes = merge_modes
        self.default_branch = default_branch
        self.branches = {}
        # The set of flags we have performed queries for:
        self.completed_flags = BranchFlag.CLEAR
        # If there was an error fetching the branches for a given set
        # of flags, the failure will be recorded here:
        self.failed_flags = BranchFlag.CLEAR

    def toDict(self):
        return {
            'merge_modes': self.merge_modes,
            'default_branch': self.default_branch,
            'branches': {b.name: b.toDict() for b in self.branches.values()},
            'flags': [
                self.completed_flags,
                self.failed_flags,
            ],
        }

    @classmethod
    def fromDict(cls, name, data):
        o = cls(name)
        o.merge_modes = data['merge_modes']
        o.default_branch = data['default_branch']
        o.branches = {
            name: BranchInfo.fromDict(name, bdata)
            for name, bdata in data['branches'].items()
        }
        o.completed_flags = BranchFlag(data['flags'][0])
        o.failed_flags = BranchFlag(data['flags'][1])
        return o


class BranchCacheZKObject(ShardedZKObject):
    """Store the branch cache in ZK

    If a project is absent from the dict, it needs to be queried from
    the source.

    If there was an error fetching the branches, None will be stored
    as a sentinel value.
    """

    # We can always recreate data if necessary, so go ahead and
    # truncate when we update so we avoid corrupted data.
    truncate_on_create = True

    def getPath(self):
        return self._path

    def __init__(self):
        super().__init__()
        self._set(
            projects={},
        )

    def serialize(self, context):
        data = {
            "projects": {p.name: p.toDict() for p in self.projects.values()},
        }
        return json.dumps(data, sort_keys=True).encode("utf8")

    def deserialize(self, raw, context, extra=None):
        data = super().deserialize(raw, context, extra)
        projects = {}
        for project_name, project_data in data['projects'].items():
            projects[project_name] = ProjectInfo.fromDict(
                project_name, project_data)
        data['projects'] = projects
        return data


class BranchCache:
    def __init__(self, zk_client, connection, component_registry):
        self.log = logging.getLogger(
            f"zuul.BranchCache.{connection.connection_name}")

        self.connection = connection

        cname = self.connection.connection_name
        base_path = f'/zuul/cache/connection/{cname}/branches'
        lock_path = f'{base_path}/lock'
        data_path = f'{base_path}/data'

        self.rlock = SessionAwareReadLock(zk_client.client, lock_path)
        self.wlock = SessionAwareWriteLock(zk_client.client, lock_path)

        # TODO: standardize on a stop event for connections and add it
        # to the context.
        self.zk_context = ZKContext(zk_client, self.wlock, None, self.log)

        with (self.zk_context as ctx,
              zk_locked(self.wlock)):
            try:
                self.cache = BranchCacheZKObject.fromZK(
                    ctx, data_path, _path=data_path)
            except NoNodeError:
                self.cache = BranchCacheZKObject.new(
                    ctx, _path=data_path)

    def clear(self, projects=None):
        """Clear the cache"""
        with (zk_locked(self.wlock),
              self.zk_context as ctx,
              self.cache.activeContext(ctx)):
            if projects is None:
                self.cache.projects.clear()
            else:
                for p in projects:
                    self.cache.projects.pop(p, None)

    def getProjectCompletedFlags(self, project_name):
        """Get the completed branch query flags for a project

        :param str project_name: The project name

        :returns: a BranchFlag of the completed query flags
        """
        try:
            project_info = self.cache.projects[project_name]
        except KeyError:
            return BranchFlag.CLEAR
        return project_info.completed_flags

    def getProjectBranches(self, project_name, required_flags,
                           min_ltime=-1, default=RAISE_EXCEPTION):
        """Get the branch names for the given project.

        Checking the branch cache we need to distinguish three different
        cases:

            1. cache miss (not queried yet)
            2. cache hit (including empty list of branches)
            3. error when fetching branches

        If the cache doesn't contain any branches for the project and no
        default value is provided a LookupError is raised.

        If there was an error fetching the branches, the return value
        will be None.

        Otherwise the list of branches will be returned.

        :param str project_name:
            The project for which the branches are returned.
        :param bool required_flags:
            The branch flags we must have completed queries for in order
            for the cache to be considered valid.
        :param int min_ltime:
            The minimum cache ltime to consider the cache valid.
        :param any default:
            Optional default value to return if no cache entry exits.

        :returns: The list of branch names, or None if there was
            an error when fetching the branches.
        """
        if self.ltime < min_ltime:
            with (zk_locked(self.rlock),
                  self.zk_context as ctx):
                self.cache.refresh(ctx)

        project_info = None
        try:
            project_info = self.cache.projects[project_name]
        except KeyError:
            return return_default(default, project_name)

        # We've definitely stored a failure, so return that.
        if project_info is None:
            return None

        # Determine if we have enough info to answer the question
        # Check that required flags is a subset of completed flags
        if not (required_flags & project_info.completed_flags
                == required_flags):
            # We don't have the data, either because we haven't
            # queried it or the query failed.  Figure out which.
            # If there is any overlap, something failed.
            if (required_flags & project_info.failed_flags):
                return None
            return return_default(default, project_name)

        # We have the necessary info for this filtering.
        return list(project_info.branches.values())

    def setProjectBranches(self, project_name,
                           valid_flags, branch_infos):
        """Set the branch names for the given project.

        Use None as a sentinel value for the branches to indicate that
        there was a fetch error.

        :param str project_name:
            The project for the branches.
        :param BranchFlag valid_flags:
            The queries this list of branches is able to satisfy.
        :param list[str] branches:
            The list of branches or None to indicate a fetch error.
        """
        with (zk_locked(self.wlock),
              self.zk_context as ctx,
              self.cache.activeContext(ctx)):
            self._setProjectBranches(project_name,
                                     valid_flags, branch_infos)

    def _setProjectBranches(self, project_name,
                            valid_flags, branch_infos):
        project_info = self.cache.projects.get(project_name)
        if project_info is None:
            project_info = ProjectInfo(project_name)
            self.cache.projects[project_name] = project_info

        if branch_infos is None:
            # We're storing an error, set the bits accordingly
            project_info.failed_flags |= valid_flags
            project_info.completed_flags &= ~valid_flags
            return

        # Set the bits indicating a good query.
        project_info.failed_flags &= ~valid_flags
        project_info.completed_flags |= valid_flags

        # Add or update branch info
        for branch_info in branch_infos:
            existing = project_info.branches.get(branch_info.name)
            if existing:
                existing.update(branch_info)
            else:
                project_info.branches[branch_info.name] = branch_info

        # Delete any existing branches which we would expect to be
        # in the results but aren't.  At the time of writing, this
        # isn't strictly necessary because we clear the branch
        # cache on branch deletion, but this may enable us to
        # change that in the future.
        valid_branches = set(bi.name for bi in branch_infos)
        for branch_name in list(project_info.branches.keys()):
            if branch_name in valid_branches:
                continue
            branch_info = project_info.branches[branch_name]
            # If the branch_info flags are a subset of the valid
            # flags, we can delete it.
            if (branch_info.valid_flags & valid_flags ==
                branch_info.valid_flags):
                del project_info.branches[branch_name]

    def setProtected(self, project_name, branch, protected):
        """Correct the protection state of a branch.

        This may be called if a branch has changed state without us
        receiving an explicit event.
        """
        with (zk_locked(self.wlock),
              self.zk_context as ctx,
              self.cache.activeContext(ctx)):

            project_info = self.cache.projects.get(project_name)
            if project_info is None:
                project_info = ProjectInfo(project_name)
                self.cache.projects[project_name] = project_info

            branch_info = project_info.branches.get(branch)
            if branch_info is None:
                branch_info = BranchInfo(branch)
                project_info.branches[branch] = branch_info

            branch_info.protected = protected

    def getProjectMergeModes(self, project_name,
                             min_ltime=-1, default=RAISE_EXCEPTION):
        """Get the merge modes for the given project.

        Checking the branch cache we need to distinguish three different
        cases:

            1. cache miss (not queried yet)
            2. cache hit (including empty list of merge modes)
            3. error when fetching merge modes

        If the cache doesn't contain any merge modes for the project and no
        default value is provided a LookupError is raised.

        If there was an error fetching the merge modes, the return value
        will be None.

        Otherwise the list of merge modes will be returned.

        :param str project_name:
            The project for which the merge modes are returned.
        :param int min_ltime:
            The minimum cache ltime to consider the cache valid.
        :param any default:
            Optional default value to return if no cache entry exits.

        :returns: The list of merge modes by model id, or None if there was
            an error when fetching the merge modes.
        """
        if self.ltime < min_ltime:
            with (zk_locked(self.rlock),
                  self.zk_context as ctx):
                self.cache.refresh(ctx)

        project_info = None
        try:
            project_info = self.cache.projects[project_name]
        except KeyError:
            return return_default(default, project_name)

        if project_info is None:
            return None

        return project_info.merge_modes

    def setProjectMergeModes(self, project_name, merge_modes):
        """Set the supported merge modes for the given project.

        Use None as a sentinel value for the merge modes to indicate
        that there was a fetch error.

        :param str project_name:
            The project for the merge modes.
        :param list[int] merge_modes:
            The list of merge modes (by model ID) or None.

        """
        with (zk_locked(self.wlock),
              self.zk_context as ctx,
              self.cache.activeContext(ctx)):
            self._setProjectMergeModes(project_name, merge_modes)

    def _setProjectMergeModes(self, project_name, merge_modes):
        project_info = self.cache.projects.get(project_name)
        if project_info is None:
            project_info = ProjectInfo(project_name)
        project_info.merge_modes = merge_modes

    def getProjectDefaultBranch(self, project_name,
                                min_ltime=-1, default=RAISE_EXCEPTION):
        """Get the default branch for the given project.

        Checking the branch cache we need to distinguish three different
        cases:

            1. cache miss (not queried yet)
            2. cache hit (including unknown default branch)
            3. error when fetching default branch

        If the cache doesn't contain a default branch for the project
        and no default value is provided a LookupError is raised.

        If there was an error fetching the default branch, the return
        value will be None.

        Otherwise the default branch will be returned.

        :param str project_name:
            The project for which the default branch is returned.
        :param int min_ltime:
            The minimum cache ltime to consider the cache valid.
        :param any default:
            Optional default value to return if no cache entry exits.

        :returns: The name of the default branch or None if there was
            an error when fetching it.

        """
        if self.ltime < min_ltime:
            with (zk_locked(self.rlock),
                  self.zk_context as ctx):
                self.cache.refresh(ctx)

        project_info = None
        try:
            project_info = self.cache.projects[project_name]
        except KeyError:
            return return_default(default, project_name)

        if project_info is None:
            return None

        return project_info.default_branch

    def setProjectDefaultBranch(self, project_name, default_branch):
        """Set the upstream default branch for the given project.

        Use None as a sentinel value for the default branch to indicate
        that there was a fetch error.

        :param str project_name:
            The project for the default branch.
        :param str default_branch:
            The default branch or None.

        """
        with (zk_locked(self.wlock),
              self.zk_context as ctx,
              self.cache.activeContext(ctx)):
            self._setProjectDefaultBranch(project_name, default_branch)

    def _setProjectDefaultBranch(self, project_name, default_branch):
        """Set the upstream default branch for the given project.

        Use None as a sentinel value for the default branch to indicate
        that there was a fetch error.

        :param str project_name:
            The project for the default branch.
        :param str default_branch:
            The default branch or None.

        """
        project_info = self.cache.projects.get(project_name)
        if project_info is None:
            project_info = ProjectInfo(project_name)
        project_info.default_branch = default_branch

    def setAllProjectData(self, project_name, valid_flags,
                          branches, merge_modes, default_branch):
        """Set all the branch info for a project at once

        :param str project_name:
            The project for the branches.
        :param BranchFlag valid_flags:
            The queries this list of branches is able to satisfy.
        :param list[str] branches:
            The list of branches or None to indicate a fetch error.
        :param list[int] merge_modes:
            The list of merge modes (by model ID) or None.
        :param str default_branch:
            The default branch or None.
        """
        # Set all three pieces of info with one lock and one active
        # context so that we only write the cache out once.
        with (zk_locked(self.wlock),
              self.zk_context as ctx,
              self.cache.activeContext(ctx)):
            if valid_flags is not None:
                self._setProjectBranches(project_name, valid_flags, branches)
            self._setProjectMergeModes(project_name, merge_modes)
            self._setProjectDefaultBranch(project_name, default_branch)

    @property
    def ltime(self):
        return self.cache._zstat.last_modified_transaction_id



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import json
import math
import threading
import urllib.parse

from zuul.lib.voluputil import (
    Constant,
    Nullable,
    Optional,
    Required,
    assemble,
    discriminate,
)
from zuul import model
from zuul.zk import zkobject
from zuul.zk.quotacache import QuotaCache
import zuul.provider.schema as provider_schema

import voluptuous as vs


class CNameMixin:

    @property
    def canonical_name(self):
        return '/'.join([
            urllib.parse.quote_plus(
                self.project_canonical_name),
            urllib.parse.quote_plus(self.name),
        ])


class BaseProviderImage(CNameMixin, metaclass=abc.ABCMeta):
    inheritable_cloud_schema = assemble(
        provider_schema.common_image,
    )
    inheritable_zuul_schema = assemble(
        provider_schema.common_image,
        provider_schema.common_image_zuul,
    )
    cloud_schema = assemble(
        provider_schema.common_image,
        provider_schema.base_image,
        vs.Schema({Required(
            'type',
            doc="""The type of image.""",
        ): vs.Any(Constant(
            'cloud',
            doc="""An existing image available in the cloud.""",
        ))}),
        doc="These are the attributes available for a cloud image.",
    )
    zuul_schema = assemble(
        provider_schema.common_image,
        provider_schema.common_image_zuul,
        provider_schema.base_image,
        vs.Schema({Required(
            'type',
            doc="""The type of image.""",
        ): vs.Any(Constant(
            'zuul',
            doc="""An image managed by Zuul.""",
        ))}),
        doc="These are the attributes available for a Zuul image.",
    )
    schema = vs.Union(
        cloud_schema, zuul_schema,
        discriminant=discriminate(
            lambda val, alt: val['type'] == alt['type'].validators[0])
    )
    inheritable_schema = assemble(
        inheritable_cloud_schema,
        inheritable_zuul_schema,
    )

    def __init__(self, image_config, provider_config):
        new_config = image_config.copy()
        self.__dict__.update(self.schema(new_config))


class BaseProviderFlavor(CNameMixin, metaclass=abc.ABCMeta):
    inheritable_schema = assemble(
        provider_schema.common_flavor,
    )
    schema = assemble(
        provider_schema.common_flavor,
        provider_schema.base_flavor,
    )

    def __init__(self, flavor_config, provider_config):
        new_config = flavor_config.copy()
        self.__dict__.update(self.schema(new_config))


class BaseProviderLabel(CNameMixin, metaclass=abc.ABCMeta):
    inheritable_schema = assemble(
        provider_schema.common_label,
    )
    schema = assemble(
        provider_schema.common_label,
        provider_schema.base_label,
    )
    image_flavor_inheritable_schema = assemble()

    def __init__(self, label_config, provider_config):
        new_config = label_config.copy()
        self.__dict__.update(self.schema(new_config))

    def __repr__(self):
        return (f"<{self.__class__.__name__} "
                f"canonical_name={self.canonical_name} >")

    def inheritFrom(self, image, flavor):
        # Some label attributes should default to values specified in
        # a flavor or image (for example, volume types) but can be
        # overridden by a label.  This method implements that
        # inheritance, using the attributes specified in
        # image_label_inhertable_schema.
        for attr in self.image_flavor_inheritable_schema.schema:
            # Get the vs.Optional attribute mutation
            if hasattr(attr, 'output'):
                attr = attr.output
            if getattr(self, attr, None) is None:
                setattr(self, attr,
                        getattr(flavor, attr, None) or
                        getattr(image, attr, None))


class BaseProviderEndpoint(metaclass=abc.ABCMeta):
    """Base class for provider endpoints.

    Providers and Sections are combined to describe clouds, and they
    may not correspond exactly with the cloud's topology.  To
    reconcile this, the Endpoint class is used for storing information
    about what we would typically call a region of a cloud.  This is
    the unit of visibility of instances, VPCs, images, etc.
    """

    def __init__(self, driver, zk_client, connection, name, system_id):
        self.driver = driver
        self.zk_client = zk_client
        self.connection = connection
        self.name = name
        self.system_id = system_id
        self.start_lock = threading.Lock()
        self.started = False
        self.stopped = False
        self.quota_cache = QuotaCache(zk_client, name)

    def start(self):
        with self.start_lock:
            if not self.stopped and not self.started:
                self.startEndpoint()
                self.started = True

    def stop(self):
        with self.start_lock:
            if self.started:
                self.stopEndpoint()
            # Set the stopped flag regardless of whether we started so
            # that we won't start after stopping.
            self.stopped = True

    @property
    def canonical_name(self):
        return '/'.join([
            urllib.parse.quote_plus(self.connection.connection_name),
            urllib.parse.quote_plus(self.name),
        ])

    def startEndpoint(self):
        """Start the endpoint

        This method may start any threads necessary for the endpoint.

        """
        raise NotImplementedError()

    def stopEndpoint(self):
        """Stop the endpoint

        This method must stop all endpoint threads.
        """
        raise NotImplementedError()

    def refreshQuotaLimits(self, update):
        """Query the endpoint for quota limits and store them in the quota
        cache

        :param bool update: Whether an update should be performed even
            if there are values present.

        :return: Whether the cache was updated
        :rtype: bool

        """
        raise NotImplementedError()

    def postConfig(self, provider):
        """Perform any endpoint-global actions after reconfiguration

        This will be called any time a tenant layout is updated, once
        for each provider that uses the endpoint.  It may be called
        multiple times for the same update, and it may be called even
        when nothing about the provider or endpoint changed.

        """
        pass

    def listInstances(self):
        """Return an iterator of instances accessible to this endpoint.

        The yielded values should represent all instances accessible
        to this endpoint, not only those under the control of Zuul,
        but all visible instances in order to achive accurate quota
        calculation.

        :returns: A generator of :py:class:`Instance` objects.

        """
        raise NotImplementedError()

    def listResources(self, providers):
        """Return a list of resources accessible to this endpoint.

        The yielded values should represent all resources accessible
        to this endpoint, not only those under the control of Zuul,
        but all visible instances in order for the driver to identify
        leaked resources and instruct the endpoint to remove them.

        :param list providers: A list of providers that share this endpoint.

        :returns: A generator of :py:class:`Resource` objects.

        """
        raise NotImplementedError()

    def listRequiredInfrastructureResources(self):
        """Return a list of infrastructure resources in use by this endpoint.

        This returns the same type of resource objects as
        listResources, but it returns only the resources which are
        currently needed by the provider.  In "infrastructure
        resource" is a resource which is not directly associated with
        a single node or upload.  Instead, it is a resource that may
        be used by multiple nodes or uploads.  Examples include launch
        templates or ssh keys.  Providers may create these on their
        own when needed by their labels.  The launcher will use this
        information to help the provider know when to delete them.

        :returns: A generator of :py:class:`Resource` objects.

        """

        return []

    def deleteResource(self, resource):

        """Delete the supplied resource

        The driver has identified a leaked resource and the adapter
        should delete it.

        :param Resource resource: A Resource object previously
            returned by 'listResources'.
        """
        raise NotImplementedError()


class BaseProviderSchema(metaclass=abc.ABCMeta):
    def getLabelSchema(self):
        return BaseProviderLabel.schema

    def getImageSchema(self):
        return BaseProviderImage.schema

    def getFlavorSchema(self):
        return BaseProviderFlavor.schema

    def getInheritableLabelSchema(self):
        return BaseProviderLabel.inheritable_schema

    def getInheritableImageSchema(self):
        return BaseProviderImage.inheritable_schema

    def getInheritableZuulImageSchema(self):
        return BaseProviderImage.inheritable_zuul_schema

    def getInheritableCloudImageSchema(self):
        return BaseProviderImage.inheritable_cloud_schema

    def getInheritableFlavorSchema(self):
        return BaseProviderFlavor.inheritable_schema

    def getProviderSchema(self):
        schema = vs.Schema({
            '_source_context': model.SourceContext,
            '_start_mark': model.ZuulMark,
            Required(
                'name',
                doc="""\
                The name of the provider.  Used to refer to the
                provider in Zuul configuration.""",
            ): str,
            Required(
                'section',
                doc="""\
                The name of the :attr:`section` from which to inherit.""",
            ): str,
            Required(
                'labels',
                doc="""A list of labels associated with this provider.""",
            ): [self.getLabelSchema()],
            Required(
                'images',
                doc="""A list of images associated with this provider.""",
            ): [self.getImageSchema()],
            Required(
                'flavors',
                doc="""A list of flavors associated with this provider.""",
            ): [self.getFlavorSchema()],
            Optional(
                'label-defaults',
                doc="""\
                Attributes to be set as default values for any label
                used with this provider.  Many attributes which may be
                set on an individual label may be set once in this
                section and used for all the labels in this provider.
                Values set on individual labels may still override the
                values set here.""",
                default={},
            ): self.getInheritableLabelSchema(),
            Optional(
                'image-defaults',
                doc="""\
                Attributes to be set as default values for any image
                used with this provider.  Many attributes which may be
                set on an individual image may be set once in this
                section and used for all the images in this provider.
                Values set on individual images may still override the
                values set here.""",
                default={},
            ): self.getInheritableImageSchema(),
            Optional(
                'flavor-defaults',
                doc="""\
                Attributes to be set as default values for any flavor
                used with this provider.  Many attributes which may be
                set on an individual flavor may be set once in this
                section and used for all the flavors in this provider.
                Values set on individual flavors may still override the
                values set here.""",
                default={},
            ): self.getInheritableFlavorSchema(),
            Optional(
                'abstract',
                doc="""\
                Whether a section is intended to be inherited by
                another :attr:`section` or a :attr:`provider`.  This
                setting is currently unused (but may be used in the
                future).  If a section is used to provide common
                values to other sections, set this to `true`.
                Otherwise, the default of `false` indicates that the
                section should be referenced directly by providers.""",
                default=False,
            ): Nullable(bool),
            Optional(
                'parent',
                doc="""\
                The name of the parent section from which to inherit.
                This attribute is only used by :attr:`section`
                objects.  To indicate which section a provider should
                be attached to, use
                :attr:`provider[$driver].section`""",
            ): Nullable(str),
            Required(
                'connection',
                doc="""\
                The name of the :ref:`connection <connections>` to
                use.  This attribute is only used by :attr:`section`
                objects and may not be changed via inheritance.""",
            ): str,
            Optional(
                'launch-timeout',
                doc="""\
                The time to wait from issuing the command to create a
                new node until the node is reporting as running.  If
                the timeout is exceeded, the node launch is aborted
                and the node deleted.""",
            ): Nullable(int),
            Optional(
                'launch-attempts',
                doc="""\
                The number of times to attempt launching a node before
                considering the request failed.""",
                default=3,
            ): int,
        })
        return schema


class BaseImageJob:
    """Abstract class to encapsulate an image creation"""

    def __init__(self):
        self.dependents = []


class BaseImageImportJob(BaseImageJob):
    """Abstract class to encapsulate an image import

    This class should contain all the information needed to perform an
    image import.  The run method will be executed asynchronously in
    an executor.
    """

    @abc.abstractmethod
    def run(self):
        """Run the import.

        :return: The external id of the image in the cloud
        """
        pass


class BaseImageUploadJob(BaseImageJob):
    """Abstract class to encapsulate an image upload

    This class should contain all the information needed to perform an
    image upload, except the filename.  The run method will be
    executed asynchronously in an executor.
    """

    @abc.abstractmethod
    def run(self, filename):
        """Run the import.

        :param filename str: Path to the local file to upload

        :return: The external id of the image in the cloud
        """
        pass


class BaseImageCopyJob(BaseImageJob):
    """Abstract class to encapsulate an image copy

    This class should contain all the information needed to perform an
    image copy.  The run method will be executed asynchronously in
    an executor.
    """

    @abc.abstractmethod
    def run(self, external_id):
        """Run the import.

        :param external_id str: The external id of the image to copy

        :return: The external id of the image in the cloud
        """
        pass


class BaseProvider(zkobject.PolymorphicZKObjectMixin,
                   zkobject.ShardedZKObject):

    """Base class for provider."""
    schema = BaseProviderSchema().getProviderSchema()

    def __init__(self, *args):
        super().__init__()
        self._set(nodes=[])
        if args:
            (driver, zk_client, connection, tenant_name,
             canonical_name, config, system_id) = args
            config = config.copy()
            config.pop('_source_context')
            config.pop('_start_mark')
            attrs = dict(
                driver=driver,
                connection=connection,
                connection_name=connection.connection_name,
                tenant_name=tenant_name,
                canonical_name=canonical_name,
                config=config,
                system_id=system_id,
                zk_client=zk_client,
            )
            # Some of the values above are needed when parsing
            # (especially images)
            self._set(**attrs)
            parsed_config = self.parseConfig(config, connection)
            parsed_config.pop('connection')
            parsed_config.update(attrs)
            self._set(**parsed_config)

    def __repr__(self):
        return (f"<{self.__class__.__name__} "
                f"canonical_name={self.canonical_name}>")

    @classmethod
    def fromZK(cls, context, path, connections, system_id, zk_client):
        """Deserialize a Provider (subclass) from ZK.

        To deserialize a Provider from ZK, pass the connection
        registry as the "connections" argument.

        The Provider subclass will automatically be deserialized and
        the connection/driver attributes updated from the connection
        registry.

        """
        raw_data, zstat = cls._loadData(context, path)
        extra = {
            'connections': connections,
            'system_id': system_id,
            'zk_client': zk_client,
        }
        obj = cls._fromRaw(context, raw_data, zstat, extra)
        connection = connections.connections[obj.connection_name]
        obj._set(
            connection=connection,
            driver=connection.driver,
            system_id=system_id,
            zk_client=zk_client,
        )
        return obj

    def getProviderSchema(self):
        return self.schema

    def parseConfig(self, config, connection):
        schema = self.getProviderSchema()
        ret = schema(config)
        images = self.parseImages(config, connection)
        flavors = self.parseFlavors(config, connection)
        labels = self.parseLabels(config, connection)
        for label in labels.values():
            label.inheritFrom(images[label.image], flavors[label.flavor])
        ret.update(dict(
            images=images,
            flavors=flavors,
            labels=labels,
        ))
        self.validateConfig(ret)
        return ret

    def deserialize(self, raw, context, extra):
        self._set(
            system_id=extra['system_id'],
            zk_client=extra['zk_client'],
        )
        data = super().deserialize(raw, context)
        connections = extra['connections']
        connection = connections.connections[data['connection_name']]
        data['connection'] = connection
        data['driver'] = connection.driver
        self._set(
            canonical_name=data['canonical_name'],
        )
        data.update(self.parseConfig(data['config'], connection))
        return data

    def serialize(self, context):
        data = dict(
            tenant_name=self.tenant_name,
            canonical_name=self.canonical_name,
            config=self.config,
            connection_name=self.connection.connection_name,
        )
        return json.dumps(data, sort_keys=True).encode("utf8")

    @property
    def tenant_scoped_name(self):
        return f'{self.tenant_name}-{self.name}'

    def parseImages(self, config, connection):
        images = {}
        for image_config in config.get('images', []):
            i = self.parseImage(image_config, config, connection)
            images[i.name] = i
        return images

    def parseFlavors(self, config, connection):
        flavors = {}
        for flavor_config in config.get('flavors', []):
            f = self.parseFlavor(flavor_config, config, connection)
            flavors[f.name] = f
        return flavors

    def parseLabels(self, config, connection):
        labels = {}
        for label_config in config.get('labels', []):
            l = self.parseLabel(label_config, config, connection)
            labels[l.name] = l
        return labels

    @abc.abstractmethod
    def parseLabel(self, label_config, provider_config):
        """Instantiate a ProviderLabel subclass

        :returns: a ProviderLabel subclass
        :rtype: ProviderLabel
        """
        pass

    @abc.abstractmethod
    def parseFlavor(self, flavor_config, provider_config):
        """Instantiate a ProviderFlavor subclass

        :returns: a ProviderFlavor subclass
        :rtype: ProviderFlavor
        """
        pass

    @abc.abstractmethod
    def parseImage(self, image_config, provider_config):
        """Instantiate a ProviderImage subclass

        :returns: a ProviderImage subclass
        :rtype: ProviderImage
        """
        pass

    @abc.abstractmethod
    def getEndpoint(self):
        """Get an endpoint for this provider"""
        pass

    def validateConfig(self, config):
        """Validate the full and final config for this provider

        This is called after all schema validation and configuration
        inheritance has been performed.  This allows us to validate
        multiple settings from different areas (eg, that a label is
        valid with a certain flavor config).

        """
        pass

    def getPath(self):
        path = (f'/zuul/tenant/{self.tenant_name}'
                f'/provider/{self.canonical_name}/config')
        return path

    def hasLabel(self, label):
        return label in self.labels

    def getNodeTags(self, system_id, label, node_uuid,
                    provider=None, request=None):
        """Return the tags that should be stored with the node

        :param str system_id: The Zuul system uuid
        :param ProviderLabel label: The node label
        :param str node_uuid: The node uuid
        :param Provider provider: The cloud provider or None
        :param NodesetRequest request: The node request or None
        """
        tags = dict()
        attrs = request.getSafeAttributes().toDict() if request else {}
        for k, v in label.tags.items():
            try:
                tags[k] = v.format(**attrs)
            except Exception:
                self.log.exception("Error formatting metadata %s", k)

        fixed = {
            'zuul_system_id': system_id,
            'zuul_node_uuid': node_uuid,
        }
        tags.update(fixed)
        return tags

    def getCreateStateMachine(self, node,
                              image_external_id,
                              log):
        """Return a state machine suitable for creating an instance

        This method should return a new state machine object
        initialized to create the described node.

        :param ProviderNode node: The node object.
        :param ProviderLabel label: A config object representing the
            provider-label for the node.
        :param str image_external_id: If provided, the external id of
            a previously uploaded image; if None, then the adapter should
            look up a cloud image based on the label.
        :param log Logger: A logger instance for emitting annotated
            logs related to the request.

        :returns: A :py:class:`StateMachine` object.

        """
        raise NotImplementedError()

    def getDeleteStateMachine(self, node, log):
        """Return a state machine suitable for deleting an instance

        This method should return a new state machine object
        initialized to delete the described instance.

        :param node ProviderNode: The node that should be deleted.
        :param log Logger: A logger instance for emitting annotated
            logs related to the request.
        """
        raise NotImplementedError()

    def getEndpointLimits(self):
        """Return the endpoint quota limits for this provider

        The default implementation returns a simple QuotaInformation
        with no limits.  Override this to provide accurate
        information.

        :returns: A :py:class:`QuotaInformation` object.

        """
        return model.QuotaInformation(default=math.inf)

    def getQuotaForLabel(self, label):
        """Return information about the quota used for a label

        The default implementation returns a simple QuotaInformation
        for one instance; override this to return more detailed
        information including cores and RAM.

        :param ProviderLabel label: A config object describing
            a label for an instance.

        :returns: A :py:class:`QuotaInformation` object.
        """
        return model.QuotaInformation(instances=1)

    def refreshQuotaForLabel(self, label, update):
        """Query the endpoint for quota used for a label and update
        the quota cache

        :param ProviderLabel label: A config object describing
            a label for an instance.
        :param bool update: Whether an update should be performed even
            if there are values present.
        """
        raise NotImplementedError()

    def getAZs(self):
        """Return a list of availability zones for this provider

        One of these will be selected at random and supplied to the
        create state machine.  If a request handler is building a node
        set from an existing ready node, then the AZ from that node
        will be used instead of the results of this method.

        :returns: A list of availability zone names.
        """
        return [None]

    def labelReady(self, label):
        """Indicate whether a label is ready in the provided cloud

        This is used by the launcher to determine whether it should
        consider a label to be in-service for a provider.  If this
        returns False, the label will be ignored for this provider.

        This does not need to consider whether a diskimage is ready;
        the launcher handles that itself.  Instead, this can be used
        to determine whether a cloud-image is available.

        :param ProviderLabel label: A config object describing a label
            for an instance.

        :returns: A bool indicating whether the label is ready.
        """
        return True

    # The following methods must be implemented only if image
    # management is supported:

    def getSnapshotStateMachine(self, node, log):
        """Return a state machine suitable for snapshotting an instance

        This method should return a new state machine object
        initialized to snapshot the described instance.

        :param node ProviderNode: The node that should be snapshotted.
        :param log Logger: A logger instance for emitting annotated
            logs related to the request.
        """
        return None

    def downloadUrl(self, url, path):
        """Attempt to download the given URL to the destination path

        If this provider is able to download URLs of the given form,
        it should attempt to do so and save the result.  If it can not
        handle the given URL, return None.

        This is an optional method that may be implemented in order to
        allow for image storage in cloud-specific storage systems.

        :param url str: The URL of the file to download
        :param path str: The local destination path

        :return: None if the provider can not handle the URL, or the
        path if it sucessfully downloaded it.

        """
        return None

    def getImageImportJob(self, url, provider_image, image_name,
                          image_format, metadata, md5, sha256):
        """Get an image import job if able

        If the provider can support a direct image import from the
        supplied URL, then return a BaseImageUploadJob that will do so.

        :param url str: The URL of the image
        :param provider_image ProviderImageConfig:
            The provider's config for this image
        :param image_name str: The name of the image
        :param image_format str: The format of the image (e.g., "qcow")
        :param metadata dict: A dictionary of metadata that must be
            stored on the image in the cloud.
        :param md5 str: The md5 hash of the image file
        :param sha256 str: The sha256 hash of the image file

        :return: A BaseImageUploadJob that will import the image

        """
        # Most drivers probably won't implement this.
        return None

    def getImageCopyJob(self, source_provider, provider_image, image_name,
                        image_format, metadata, md5, sha256):
        """Get an image copy job if able

        If the provider can support copying an existing image from the
        supplied provider, then return a BaseImageUploadJob that will do so.

        The external_id will be passed as an argument to the run
        method of the resulting job.

        :param source_provider Provider: The provider of the source image
        :param provider_image ProviderImageConfig:
            The provider's config for this image
        :param image_name str: The name of the image
        :param image_format str: The format of the image (e.g., "qcow")
        :param metadata dict: A dictionary of metadata that must be
            stored on the image in the cloud.
        :param md5 str: The md5 hash of the image file
        :param sha256 str: The sha256 hash of the image file

        :return: A BaseImageUploadJob that will import the image

        """
        # Most drivers probably won't implement this.
        return None

    def getImageUploadJob(self, provider_image, image_name,
                          image_format, metadata, md5, sha256):
        """Get an image upload job

        Return a BaseImageUploadJob that will upload the local filename
        to the provider as an image.

        The path to the local file to be uploaded will be passed as an
        argument to the run method of the resulting job.

        :param provider_image ProviderImageConfig:
            The provider's config for this image
        :param image_name str: The name of the image
        :param image_format str: The format of the image (e.g., "qcow")
        :param metadata dict: A dictionary of metadata that must be
            stored on the image in the cloud.
        :param md5 str: The md5 hash of the image file
        :param sha256 str: The sha256 hash of the image file

        :return: A BaseImageUploadJob that will upload the image

        """
        # Required if the driver handles images at all.
        raise NotImplementedError()

    def deleteImage(self, external_id):
        """Delete an image from the cloud

        :param external_id str: The external id of the image to delete
        """
        raise NotImplementedError()

    # The following methods are optional
    def getConsoleLog(self, label, node):
        """Return the console log from the specified server

        :param label ConfigLabel: The label config for the node
        :param ProviderNode node: The node of the server
        """
        raise NotImplementedError()

    def notifyNodescanFailure(self, label, node):
        """Notify the adapter of a nodescan failure

        :param label ConfigLabel: The label config for the node
        :param ProviderNode node: The node of the server
        """
        pass

    def postConfig(self, update=False):
        """Perform any provider-global actions after reconfiguration

        This will be called any time a tenant layout is updated, once
        for each provider.  It may be called multiple times for the
        same update, and it may be called even when nothing about the
        provider changed.

        """
        if update:
            for label in self.labels.values():
                self.refreshQuotaForLabel(label, update)

    def canReuseNode(self, node):
        """The provider can veto the reuse of a node

        If the provider decides that a node should not be reused,
        return False here.  This will only be called if the launcher
        has decided that it should be reused (so label configuration
        will already have been checked).

        :param ProviderNode node: The node
        """
        if node.state == node.State.FAILED:
            return False
        return True


class EndpointCacheMixin:
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.endpoints = {}
        self.endpoints_lock = threading.Lock()

    def getEndpointById(self, endpoint_id, create_args):
        with self.endpoints_lock:
            try:
                return self.endpoints[endpoint_id]
            except KeyError:
                pass
            endpoint = self._endpoint_class(*create_args)
            self.endpoints[endpoint_id] = endpoint
        return endpoint

    def getEndpoints(self):
        with self.endpoints_lock:
            return list(self.endpoints.values())

    def stopEndpoints(self):
        with self.endpoints_lock:
            for endpoint in self.endpoints.values():
                endpoint.stop()



# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import os.path
from urllib.parse import quote_plus

import zuul.model


def unique_project_name(project_name):
    parts = project_name.split('/')
    prefix = parts[0]
    name = quote_plus(project_name)
    return (prefix, name)


def safe_tenant_name(tenant_name):
    return quote_plus(tenant_name)


def workspace_project_path(hostname, project_name, scheme):
    """Return the project path based on the specified scheme"""
    if scheme == zuul.model.SCHEME_UNIQUE:
        prefix, project_name = unique_project_name(project_name)
        return os.path.join(hostname, prefix, project_name)
    elif scheme == zuul.model.SCHEME_GOLANG:
        return os.path.join(hostname, project_name)
    elif scheme == zuul.model.SCHEME_FLAT:
        parts = project_name.split('/')
        return os.path.join(parts[-1])


def b64encode(string):
    # Return a base64 encoded string (the module operates on bytes)
    return base64.b64encode(string.encode('utf8')).decode('utf8')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tests.base import (
    ZuulTestCase,
    simple_layout,
)


class TestSerial(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    @simple_layout('layouts/serial.yaml')
    def test_deploy_window(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.setMerged()
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.setMerged()
        self.fake_gerrit.addEvent(B.getChangeMergedEvent())
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 2)
        self.assertTrue(self.builds[0].hasChanges(A))
        self.assertTrue(self.builds[1].hasChanges(A))
        self.assertFalse(self.builds[0].hasChanges(B))
        self.assertFalse(self.builds[1].hasChanges(B))

        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 2)
        self.assertTrue(self.builds[0].hasChanges(A))
        self.assertTrue(self.builds[1].hasChanges(A))
        self.assertTrue(self.builds[0].hasChanges(B))
        self.assertTrue(self.builds[1].hasChanges(B))

        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 1)
        self.assertHistory([
            dict(name='job1', result='SUCCESS', changes='1,1'),
            dict(name='job2', result='SUCCESS', changes='1,1'),
            dict(name='job1', result='SUCCESS', changes='2,1'),
            dict(name='job2', result='SUCCESS', changes='2,1'),
        ], ordered=False)

    @simple_layout('layouts/serial.yaml')
    def test_deploy_shared(self):
        # Same as test_deploy_window but with two separate projects
        # sharing a queue.
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        A.setMerged()
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        B.setMerged()
        self.fake_gerrit.addEvent(B.getChangeMergedEvent())
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 1)
        self.assertTrue(self.builds[0].hasChanges(A))

        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 1)
        self.assertTrue(self.builds[0].hasChanges(B))

        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 1)
        self.assertHistory([
            dict(name='job1', result='SUCCESS', changes='1,1'),
            dict(name='job1', result='SUCCESS', changes='2,1'),
        ], ordered=False)

    @simple_layout('layouts/serial.yaml')
    def test_deploy_unshared(self):
        # Test two projects which don't share a queue.
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.setMerged()
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()

        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'B')
        B.setMerged()
        self.fake_gerrit.addEvent(B.getChangeMergedEvent())
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 3)
        self.assertTrue(self.builds[0].hasChanges(A))
        self.assertTrue(self.builds[1].hasChanges(A))
        self.assertTrue(self.builds[2].hasChanges(B))
        self.assertFalse(self.builds[2].hasChanges(A))

        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 1)
        self.assertHistory([
            dict(name='job1', result='SUCCESS', changes='1,1'),
            dict(name='job2', result='SUCCESS', changes='1,1'),
            dict(name='job1', result='SUCCESS', changes='2,1'),
        ], ordered=False)



# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import kazoo.recipe.lock


class SemaphoreCleanupLock(kazoo.recipe.lock.Lock):
    _path = '/zuul/cleanup/sempahores'

    def __init__(self, client):
        super().__init__(client.client, self._path)


class BuildRequestCleanupLock(kazoo.recipe.lock.Lock):
    _path = '/zuul/cleanup/build_requests'

    def __init__(self, client):
        super().__init__(client.client, self._path)


class MergeRequestCleanupLock(kazoo.recipe.lock.Lock):
    _path = '/zuul/cleanup/merge_requests'

    def __init__(self, client):
        super().__init__(client.client, self._path)


class ConnectionCleanupLock(kazoo.recipe.lock.Lock):
    _path = '/zuul/cleanup/connection'

    def __init__(self, client):
        super().__init__(client.client, self._path)


class GeneralCleanupLock(kazoo.recipe.lock.Lock):
    _path = '/zuul/cleanup/general'

    def __init__(self, client):
        super().__init__(client.client, self._path)


class NodeRequestCleanupLock(kazoo.recipe.lock.Lock):
    _path = '/zuul/cleanup/node_request'

    def __init__(self, client):
        super().__init__(client.client, self._path)



#!/usr/bin/env python3
# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import requests


def main():
    parser = argparse.ArgumentParser(
        description="Find where a project declares a queue")
    parser.add_argument("url", help="Zuul URL")
    parser.add_argument("tenant", help="Zuul tenant name")
    parser.add_argument("--verbose", help="Display progress",
                        action='store_true')
    args = parser.parse_args()

    projects = requests.get(
        f'{args.url}/api/tenant/{args.tenant}/projects',
    ).json()

    pipeline_contexts = set()
    for tenant_project in projects:
        if args.verbose:
            print(f"Checking {tenant_project['name']}")
        project = requests.get(
            f"{args.url}/api/tenant/{args.tenant}/project/"
            f"{tenant_project['name']}",
        ).json()

        for config in project['configs']:
            for pipeline in config['pipelines']:
                if pipeline['queue_name']:
                    pipeline_contexts.add(repr(config['source_context']))

    if pipeline_contexts:
        print("The following project-pipeline stanzas define a queue.")
        print("This syntax is deprecated and queue definitions should")
        print("be moved to the project level.")
        print("See https://zuul-ci.org/docs/zuul/latest/"
              "releasenotes.html#relnotes-4-1-0-deprecation-notes")
        for c in pipeline_contexts:
            print(c)
    else:
        print("Good, no project-pipeline queue definitions found.")


if __name__ == '__main__':
    main()



# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tests.base import (
    ZuulTestCase,
    simple_layout,
    skipIfMultiScheduler,
)


class TestGerritLegacyCRD(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_crd_gate(self):
        "Test cross-repo dependencies"
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)

        AM2 = self.fake_gerrit.addFakeChange('org/project1', 'master', 'AM2')
        AM1 = self.fake_gerrit.addFakeChange('org/project1', 'master', 'AM1')
        AM2.setMerged()
        AM1.setMerged()

        BM2 = self.fake_gerrit.addFakeChange('org/project2', 'master', 'BM2')
        BM1 = self.fake_gerrit.addFakeChange('org/project2', 'master', 'BM1')
        BM2.setMerged()
        BM1.setMerged()

        # A -> AM1 -> AM2
        # B -> BM1 -> BM2
        # A Depends-On: B
        # M2 is here to make sure it is never queried.  If it is, it
        # means zuul is walking down the entire history of merged
        # changes.

        B.setDependsOn(BM1, 1)
        BM1.setDependsOn(BM2, 1)

        A.setDependsOn(AM1, 1)
        AM1.setDependsOn(AM2, 1)

        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])

        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(B.data['status'], 'NEW')

        for connection in self.scheds.first.connections.connections.values():
            connection.maintainCache([], max_age=0)

        self.executor_server.hold_jobs_in_build = True
        B.addApproval('Approved', 1)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(AM2.queried, 0)
        self.assertEqual(BM2.queried, 0)
        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(A.reported, 2)
        self.assertEqual(B.reported, 2)

        changes = self.getJobFromHistory(
            'project-merge', 'org/project1').changes
        self.assertEqual(changes, '2,1 1,1')

    def test_crd_branch(self):
        "Test cross-repo dependencies in multiple branches"

        self.create_branch('org/project2', 'mp')
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        C1 = self.fake_gerrit.addFakeChange('org/project2', 'mp', 'C1')
        C2 = self.fake_gerrit.addFakeChange('org/project2', 'mp', 'C2',
                                            status='ABANDONED')
        C1.data['id'] = B.data['id']
        C2.data['id'] = B.data['id']

        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        C1.addApproval('Code-Review', 2)

        # A Depends-On: B+C1
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])

        self.executor_server.hold_jobs_in_build = True
        B.addApproval('Approved', 1)
        C1.addApproval('Approved', 1)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(C1.data['status'], 'MERGED')
        self.assertEqual(A.reported, 2)
        self.assertEqual(B.reported, 2)
        self.assertEqual(C1.reported, 2)

        changes = self.getJobFromHistory(
            'project-merge', 'org/project1').changes
        self.assertEqual(changes, '2,1 3,1 1,1')

    def test_crd_multiline(self):
        "Test multiple depends-on lines in commit"
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project2', 'master', 'C')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        C.addApproval('Code-Review', 2)

        # A Depends-On: B+C
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\nDepends-On: %s\n' % (
            A.subject, B.data['id'], C.data['id'])

        self.executor_server.hold_jobs_in_build = True
        B.addApproval('Approved', 1)
        C.addApproval('Approved', 1)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(C.data['status'], 'MERGED')
        self.assertEqual(A.reported, 2)
        self.assertEqual(B.reported, 2)
        self.assertEqual(C.reported, 2)

        changes = self.getJobFromHistory(
            'project-merge', 'org/project1').changes
        self.assertEqual(changes, '2,1 3,1 1,1')

    def test_crd_unshared_gate(self):
        "Test cross-repo dependencies in unshared gate queues"
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])

        # A and B do not share a queue, make sure that A is unable to
        # enqueue B (and therefore, A is unable to be enqueued).
        B.addApproval('Approved', 1)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 0)
        self.assertEqual(len(self.history), 0)

        # Enqueue and merge B alone.
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(B.reported, 2)

        # Now that B is merged, A should be able to be enqueued and
        # merged.
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(A.reported, 3)

    def test_crd_gate_reverse(self):
        "Test reverse cross-repo dependencies"
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)

        # A Depends-On: B

        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])

        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(B.data['status'], 'NEW')

        self.executor_server.hold_jobs_in_build = True
        A.addApproval('Approved', 1)
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(A.reported, 2)
        self.assertEqual(B.reported, 2)

        changes = self.getJobFromHistory(
            'project-merge', 'org/project1').changes
        self.assertEqual(changes, '2,1 1,1')

    def test_crd_cycle(self):
        "Test cross-repo dependency cycles"
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        B.addApproval('Approved', 1)

        # A -> B -> A (via commit-depends)

        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])
        B.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            B.subject, A.data['id'])

        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 0)
        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(B.data['status'], 'NEW')

    def test_crd_gate_unknown(self):
        "Test unknown projects in dependent pipeline"
        self.init_repo("org/unknown", tag='init')
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/unknown', 'master', 'B')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])

        B.addApproval('Approved', 1)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        # Unknown projects cannot share a queue with any other
        # since they don't have common jobs with any other (they have no jobs).
        # Changes which depend on unknown project changes
        # should not be processed in dependent pipeline
        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 0)
        self.assertEqual(len(self.history), 0)

        # Simulate change B being gated outside this layout Set the
        # change merged before submitting the event so that when the
        # event triggers a gerrit query to update the change, we get
        # the information that it was merged.
        B.setMerged()
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Now that B is merged, A should be able to be enqueued and
        # merged.
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(A.reported, 3)
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(B.reported, 0)

    def test_crd_check(self):
        "Test cross-repo dependencies in independent pipelines"

        self.executor_server.hold_jobs_in_build = True
        self.hold_jobs_in_queue = True
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.hold_jobs_in_queue = False
        self.executor_api.release()
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()

        self.assertTrue(self.builds[0].hasChanges(A, B))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 0)

        self.assertEqual(self.history[0].changes, '2,1 1,1')
        self.assertEqual(len(self.getAllQueues('tenant-one', 'check')), 0)

    def test_crd_check_git_depends(self):
        "Test single-repo dependencies in independent pipelines"
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'B')

        # Add two git-dependent changes and make sure they both report
        # success.
        B.setDependsOn(A, 1)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 1)

        self.assertEqual(self.history[0].changes, '1,1')
        self.assertEqual(self.history[-1].changes, '1,1 2,1')
        self.assertEqual(len(self.getAllQueues('tenant-one', 'check')), 0)

        self.assertIn('Build succeeded', A.messages[0])
        self.assertIn('Build succeeded', B.messages[0])

    def test_crd_check_duplicate(self):
        "Test duplicate check in independent pipelines"
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'B')

        # Add two git-dependent changes...
        B.setDependsOn(A, 1)
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.getAllItems('tenant-one', 'check')), 2)

        # ...make sure the live one is not duplicated...
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.getAllItems('tenant-one', 'check')), 2)

        # ...but the non-live one is able to be.
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.getAllItems('tenant-one', 'check')), 3)

        # Release jobs in order to avoid races with change A jobs
        # finishing before change B jobs.
        self.orderedRelease()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 1)

        self.assertEqual(self.history[0].changes, '1,1 2,1')
        self.assertEqual(self.history[1].changes, '1,1')
        self.assertEqual(len(self.getAllQueues('tenant-one', 'check')), 0)

        self.assertIn('Build succeeded', A.messages[0])
        self.assertIn('Build succeeded', B.messages[0])

    def _test_crd_check_reconfiguration(self, project1, project2):
        "Test cross-repo dependencies re-enqueued in independent pipelines"

        self.hold_jobs_in_queue = True
        A = self.fake_gerrit.addFakeChange(project1, 'master', 'A')
        B = self.fake_gerrit.addFakeChange(project2, 'master', 'B')

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        # Make sure the items still share a change queue, and the
        # first one is not live.
        self.assertEqual(len(self.getAllQueues('tenant-one', 'check')), 1)
        queue = self.getAllQueues('tenant-one', 'check')[0]
        first_item = queue.queue[0]
        for item in queue.queue:
            self.assertEqual(item.queue, first_item.queue)
        self.assertFalse(first_item.live)
        self.assertTrue(queue.queue[1].live)

        self.hold_jobs_in_queue = False
        self.executor_api.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 0)

        self.assertEqual(self.history[0].changes, '2,1 1,1')
        self.assertEqual(len(self.getAllQueues('tenant-one', 'check')), 0)

    @skipIfMultiScheduler()
    def test_crd_check_reconfiguration(self):
        self._test_crd_check_reconfiguration('org/project1', 'org/project2')

    @skipIfMultiScheduler()
    def test_crd_undefined_project(self):
        """Test that undefined projects in dependencies are handled for
        independent pipelines"""
        # It's a hack for fake gerrit,
        # as it implies repo creation upon the creation of any change
        self.init_repo("org/unknown", tag='init')
        self._test_crd_check_reconfiguration('org/project1', 'org/unknown')

    @simple_layout('layouts/ignore-dependencies.yaml')
    def test_crd_check_ignore_dependencies(self):
        "Test cross-repo dependencies can be ignored"

        self.hold_jobs_in_queue = True
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project2', 'master', 'C')

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])
        # C git-depends on B
        C.setDependsOn(B, 1)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # Make sure none of the items share a change queue, and all
        # are live.
        self.assertEqual(len(self.getAllQueues('tenant-one', 'check')), 3)
        self.assertEqual(len(self.getAllItems('tenant-one', 'check')), 3)
        for item in self.getAllItems('tenant-one', 'check'):
            self.assertTrue(item.live)

        self.hold_jobs_in_queue = False
        self.executor_api.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(C.data['status'], 'NEW')
        self.assertEqual(A.reported, 1)
        self.assertEqual(B.reported, 1)
        self.assertEqual(C.reported, 1)

        # Each job should have tested exactly one change
        for job in self.history:
            self.assertEqual(len(job.changes.split()), 1)

    @simple_layout('layouts/three-projects.yaml')
    def test_crd_check_transitive(self):
        "Test transitive cross-repo dependencies"
        # Specifically, if A -> B -> C, and C gets a new patchset and
        # A gets a new patchset, ensure the test of A,2 includes B,1
        # and C,2 (not C,1 which would indicate stale data in the
        # cache for B).
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project3', 'master', 'C')

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])

        # B Depends-On: C
        B.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            B.subject, C.data['id'])

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '3,1 2,1 1,1')

        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '3,1 2,1')

        self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '3,1')

        C.addPatchset()
        self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(2))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '3,2')

        A.addPatchset()
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(2))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '3,2 2,1 1,2')

    def test_crd_check_unknown(self):
        "Test unknown projects in independent pipeline"
        self.init_repo("org/unknown", tag='init')
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/unknown', 'master', 'D')
        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])

        # Make sure zuul has seen an event on B.
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(A.reported, 1)
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(B.reported, 0)

    def test_crd_cycle_join(self):
        "Test an updated change creates a cycle"
        A = self.fake_gerrit.addFakeChange('org/project2', 'master', 'A')

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(A.reported, 1)

        # Create B->A
        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'B')
        B.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            B.subject, A.data['id'])
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # Dep is there so zuul should have reported on B
        self.assertEqual(B.reported, 1)

        # Update A to add A->B (a cycle).
        A.addPatchset()
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.data['id'])
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(2))
        self.waitUntilSettled()

        # Dependency cycle injected so zuul should have reported again on A
        self.assertEqual(A.reported, 2)

        # Now if we update B to remove the depends-on, everything
        # should be okay.  B; A->B

        B.addPatchset()
        B.data['commitMessage'] = '%s\n' % (B.subject,)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(2))
        self.waitUntilSettled()

        # Cycle was removed so now zuul should have reported again on A
        self.assertEqual(A.reported, 3)

        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(2))
        self.waitUntilSettled()
        self.assertEqual(B.reported, 2)


class TestGerritLegacyCRDWeb(TestGerritLegacyCRD):
    config_file = 'zuul-gerrit-web.conf'



# Copyright 2022-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import math

import zuul.provider.schema as provider_schema
from zuul.lib.voluputil import (
    AsList,
    Constant,
    Exclusive,
    Nullable,
    Optional,
    Required,
    RequiredExclusive,
    assemble,
    discriminate,
)

import voluptuous as vs

from zuul.driver.aws.awsendpoint import (
    AwsCreateStateMachine,
    AwsDeleteStateMachine,
    AwsSnapshotStateMachine,
)
from zuul.driver.aws.const import (
    SPOT,
    ON_DEMAND,
    VOLUME_QUOTA_CODES,
    ALL_QUOTA_CODES,
)
from zuul.model import QuotaInformation
from zuul.provider import (
    BaseProvider,
    BaseProviderFlavor,
    BaseProviderImage,
    BaseProviderLabel,
    BaseProviderSchema,
)

URL_BOTO_DESCRIBE_IMAGES = 'https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_images'  # noqa
URL_EBS_VOLUME_TYPE = 'https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/EBSVolumeTypes.html'  # noqa
URL_EBS_DIRECT_API = 'https://docs.aws.amazon.com/ebs/latest/userguide/ebs-accessing-snapshot.html'  # noqa
URL_REGISTER_IMAGE = 'https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_RegisterImage.html'  # noqa
URL_AWS_REGION = 'https://docs.aws.amazon.com/general/latest/gr/rande.html'  # noqa


class AwsProviderImage(BaseProviderImage):
    aws_image_filters = {
        Optional(
            'name',
            doc=f"""\
            The filter name. See `Boto describe images`_ for a list of
            valid filters.

            .. _`Boto describe images`: {URL_BOTO_DESCRIBE_IMAGES}
            """,
        ): Nullable(str),
        Optional(
            'values',
            doc="""A list of string values on which to filter.""",
        ): Nullable(AsList(str)),
    }
    # This is used here and in flavors and labels
    inheritable_aws_image_schema = assemble(
        vs.Schema({
            Optional(
                'volume-size',
                doc="""\
                The size of the root EBS volume, in GiB, for the
                image.  If omitted, the volume size reported for the
                imported snapshot will be used.  Only used with the
                :value:`provider[aws].images[zuul].import-method.snapshot` or
                :value:`provider[aws].images[zuul].import-method.ebs-direct`
                import methods.""",
            ): Nullable(int),
            Optional(
                'volume-type',
                doc=f"""\
                The root `EBS volume type`_ for the image.
                Only used with the
                :value:`provider[aws].images[zuul].import-method.snapshot` or
                :value:`provider[aws].images[zuul].import-method.ebs-direct`
                import methods.

                .. _`EBS volume type`: {URL_EBS_VOLUME_TYPE}
                """,
                default='gp3',
            ): str,
            Optional(
                'iops',
                doc="""\
                The number of I/O operations per second to be provisioned for
                the volume.  The default varies based on the volume type; see
                the documentation under `EBS volume type`_ for the specific
                volume type for details.
                """,
            ): Nullable(int),
            Optional(
                'throughput',
                doc="""\
                The throughput of the volume in MiB/s.  This is only valid for
                ``gp3`` volumes.
                """,
            ): Nullable(int),
            Optional(
                'imds-http-tokens',
                doc="""\
                Specify whether IMDSv2 is required.  If this is omitted,
                then AWS defaults are used (usually equivalent to
                ``optional`` but may be influenced by the image used).
                """,
            ): Nullable(vs.Any(
                Constant(
                    'optional',
                    doc="""\
                    Allows usage of IMDSv2 but do not require it.  This
                    sets the following metadata options:

                    * `HttpTokens` is `optional`
                    * `HttpEndpoint` is `enabled`
                    """),
                Constant(
                    'required',
                    doc="""\
                    Require IMDSv2.  This sets the following metadata
                    options:

                    * `HttpTokens` is `required`
                    * `HttpEndpoint` is `enabled`
                    """),
            )),
        }),
        provider_schema.cloud_image,
    )
    aws_cloud_schema = vs.Schema({
        Exclusive(
            Required(
                'image-id',
                doc="""\
                If this is provided, it is used to select the AMI
                from AWS by ID.  Either this field or
                :attr:`provider[aws].images[cloud].image-filters`
                must be provided.
                """,
            ), 'spec'): str,
        Exclusive(
            Required(
                'image-filters',
                doc="""\
                If provided, this is used to select an AMI by filters.  If
                the filters provided match more than one image, the most
                recent will be returned.  Either this field or
                :attr:`provider[aws].images[cloud].image-id` must be
                provided.

                This field may be provided as a dictionary, or a list
                of dictionaries with the following keys:
                """,
            ), 'spec'): AsList(aws_image_filters),
    })
    cloud_schema = vs.All(
        assemble(
            BaseProviderImage.cloud_schema,
            aws_cloud_schema,
            inheritable_aws_image_schema,
        ),
        RequiredExclusive('image_id', 'image_filters',
                          msg=('Provide either '
                               '"image-filters", or "image-id" keys'))
    )
    inheritable_aws_zuul_schema = vs.Schema({
        Optional(
            'import-method',
            doc="""The method to use when importing the image.""",
            default='snapshot'
        ): vs.Any(
            Constant(
                'snapshot',
                doc="""\
                This method uploads the image file to AWS as a snapshot
                and then registers an AMI directly from the snapshot.
                This is faster compared to the `image` method and may be
                used with operating systems and versions that AWS does not
                otherwise support.  However, it is incompatible with some
                operating systems which require special licensing or other
                metadata in AWS.
                """,
            ),
            Constant(
                'image',
                doc="""\
                This method uploads the image file to AWS and performs
                an "image import" on the file.  This causes AWS to
                boot the image in a temporary VM and then take a
                snapshot of that VM which is then used as the basis of
                the AMI.  This is slower compared to the `snapshot`
                method and may only be used with operating systems and
                versions which AWS already supports.  This may be
                necessary in order to use Windows images.
                """,
            ),
            Constant(
                'ebs-direct',
                doc=f"""\
                This is similar to the `snapshot` method, but uses the
                `EBS direct API`_ instead of S3.  This may be faster and
                more efficient, but it may incur additional costs.

                .. _`EBS direct API`: {URL_EBS_DIRECT_API}
                """,
            ),
        ),
        Optional(
            'image-format',
            doc="""\
            The image format that should be used when building and
            uploading or importing the image.
            """,
            default='raw',
        ): vs.Any(
            Constant(
                'ova',
                doc="""The OVA image format.""",
            ),
            Constant(
                'vhd',
                doc="""The VHD image format.""",
            ),
            Constant(
                'vhdx',
                doc="""The VHDX image format.""",
            ),
            Constant(
                'vmdk',
                doc="""The VMDK image format.""",
            ),
            Constant(
                'raw',
                doc="""A raw image.""",
            ),
            Constant(
                'snapshot',
                doc="""\
                Rather than producing an image artifact and
                uploading or importing it, this image is created by
                snapshotting a running instance.
                """,
            )
        ),
        # None is an acceptable explicit value for imds-support
        Optional(
            'imds-support',
            doc="""\
            Controls the usage of IMDSv2 on instances created from the
            image, This is only supported using the
            :value:`provider[aws].images[zuul].import-method.snapshot` or
            :value:`provider[aws].images[zuul].import-method.ebs-direct`
            import methods.
            """,
            default=None
        ): vs.Any(
            Constant(
                'v2.0',
                doc="""\
                Enforces usage of IMDSv2 by default on instances
                created from the image.
                """,
            ),
            Constant(
                None,
                doc="""IMDSv2 is optional.""",
            ),
        ),
        Optional(
            'architecture',
            doc=f"""\
            The architecture of the image.  See the `AWS RegisterImage API
            documentation`_ for valid values.

            .. _`AWS RegisterImage API documentation`: {URL_REGISTER_IMAGE}
            """,
            default='x86_64'): str,
        Optional(
            'ena-support',
            doc="""\
            Whether the image has support for the AWS Enhanced Networking
            Adapter (ENA).  Many newer operating systems include driver
            support as standard and some AWS instance types require it.
            """,
            default=True): bool,
    })
    zuul_schema = assemble(
        BaseProviderImage.zuul_schema,
        inheritable_aws_image_schema,
        inheritable_aws_zuul_schema,
    )
    inheritable_cloud_schema = assemble(
        BaseProviderImage.inheritable_cloud_schema,
        inheritable_aws_image_schema,
    )
    inheritable_zuul_schema = assemble(
        BaseProviderImage.inheritable_zuul_schema,
        inheritable_aws_image_schema,
        inheritable_aws_zuul_schema,
    )
    schema = vs.Union(
        cloud_schema, zuul_schema,
        discriminant=discriminate(
            lambda val, alt: val['type'] == alt['type'].validators[0])
    )
    inheritable_schema = assemble(
        inheritable_cloud_schema,
        inheritable_zuul_schema,
    )

    def __init__(self, image_config, provider_config):
        self.image_id = None
        self.image_filters = None
        super().__init__(image_config, provider_config)
        self.format = getattr(self, 'image_format', None)
        # Implement provider defaults
        if self.connection_type is None:
            self.connection_type = 'ssh'
        if self.connection_port is None:
            self.connection_port = 22


class AwsProviderFlavor(BaseProviderFlavor):
    fleet_schema = vs.Schema({
        Required(
            'instance-types',
            doc="""\
            A list of instance types from which AWS may select when
            launching the instance.""",
        ): AsList(str),
        Required(
            'allocation-strategy',
            doc="""\
            The allocation strategy to use when launching the instance.
            """,
        ): vs.Any(
            Constant(
                'prioritized',
                doc="""\
                The prioritized allocation strategy.  Available for
                on-demand instances""",
            ),
            Constant(
                'price-capacity-optimized',
                doc="""\
                The price-capacity-optimized allocation strategy.
                Available for spot instances""",
            ),
            Constant(
                'capacity-optimized',
                doc="""\
                The capacity-optimized allocation strategy.  Available
                for spot instances""",
            ),
            Constant(
                'diversified',
                doc="""\
                The diversified allocation strategy.  Available for
                spot instances""",
            ),
            Constant(
                'lowest-price',
                doc="""\
                The lowest-price allocation strategy.  Available for
                spot or on-demand instances""",
            ),
        )
    })

    aws_flavor_schema = vs.Schema({
        Exclusive(Required(
            'instance-type',
            doc="""\
            Name of the AWS instance type to use..
            Mutually exclusive with :attr:`provider[aws].flavors.fleet`.
            """,
        ), 'instance'): str,
        Optional(
            'dedicated-host',
            doc="""\
            If set to ``true``, an AWS dedicated host will be
            allocated for the instance.  The host may be used for one
            or more nodes depending on the settings of
            :attr:`provider[aws].labels.slots` and
            :attr:`provider[aws].labels.reuse`.

            If this option is set, the
            :value:`provider[aws].flavors.market-type.spot` option is
            not available, and :attr:`provider[aws].labels.az` option is
            required.""",
            default=False): bool,
        Optional(
            'ebs-optimized',
            doc="""\
            Indicates whether EBS optimization (additional, dedicated
            throughput between Amazon EC2 and Amazon EBS) should be
            enabled for the instance.""",
            default=False): bool,
        Optional(
            'market-type',
            doc="""\
            Whether to request an on-demand or spot instance.
            """,
            default='on-demand'
        ): vs.Any(
            Constant(
                'on-demand',
                doc="""\
                This is the typical EC2 instance where continued
                availability is guaranteed after allocation.""",
            ),
            Constant(
                'spot',
                doc="""\
                Request an Amazon EC2 Spot instance instead of an
                On-Demand instance.  Spot instances take advantage of
                unused EC2 capacity at a discount, but if demand is
                high, instances may be unavailable.  In addition,
                Amazon EC2 may interrupt Spot instances and reclaim
                them.  Alternative nodesets with On-Demand instances
                configured as a fallback may be configured in order to
                mitigate this.""",
            ),
        ),
        Exclusive(Required(
            'fleet',
            doc="""\
            If specified, the EC2 Fleet API will be used for launching
            the instance.  In this case, quota is not checked before
            launching the instance, but is taken into account after
            the instance is launched.  Mutually exclusive with
            :attr:`provider[aws].flavors.instance-type`.
            """,
        ), 'instance'): fleet_schema,
    })

    inheritable_schema = assemble(
        BaseProviderFlavor.inheritable_schema,
        # This is already included via the image, but listed again
        # here for clarity.
        AwsProviderImage.inheritable_aws_image_schema,
        provider_schema.cloud_flavor,
    )
    schema = vs.All(
        assemble(
            BaseProviderFlavor.schema,
            provider_schema.cloud_flavor,
            AwsProviderImage.inheritable_aws_image_schema,
            aws_flavor_schema,
        ),
        RequiredExclusive('instance_type', 'fleet',
                          msg=('Provide either '
                               '"instance-type", or "fleet" keys'))
    )

    def __init__(self, flavor_config, provider_config):
        self.instance_type = None
        self.fleet = None
        super().__init__(flavor_config, provider_config)


class AwsProviderLabel(BaseProviderLabel):
    aws_iam_schema = vs.All(
        vs.Schema({
            Exclusive(Required(
                'name',
                doc="""\
                Name of the instance profile.  Mutually exclusive with
                :attr:`provider[aws].labels.iam-instance-profile.arn`
                """,
            ), 'iam'): str,
            Exclusive(Required(
                'arn',
                doc="""\
                ARN identifier of the profile.  Mutually exclusive
                with
                :attr:`provider[aws].labels.iam-instance-profile.name`
                """,
            ), 'iam'): str
        }),
        RequiredExclusive('name', 'arn',
                          msg=('Provide either "name", or "arn" keys'))
    )
    aws_label_schema = vs.Schema({
        Optional(
            'az',
            doc="""\
            Instances will be assigned to the specified availibility
            zone.  If omitted, AWS will select from the available
            zones.
            """,
        ): Nullable(str),
        Optional(
            'security-group-ids',
            doc="""\
            Specifies the security group IDs to assign to the node's
            network interfaces.
            """,
        ): Nullable(AsList(str)),
        Optional(
            'subnet-ids',
            doc="""\
            Specifies the subnets to assign to the node's network interfaces.
            """,
            default=[]): AsList(str),
        Optional(
            'iam-instance-profile',
            doc="""\
            Used to attach an IAM instance profile.
            Useful for giving access to services without needing any secrets.
            """,
        ): Nullable(aws_iam_schema),
        # This could make sense for image as well, but currently is
        # only included in label.
        Optional(
            'kms-key-id',
            doc="""\
            The KMS key id to use when launching instances from an
            encrypted image.  Typically only necessary when using
            images shared from a different AWS account.
            """,
        ): Nullable(str),
    })

    inheritable_schema = assemble(
        BaseProviderLabel.inheritable_schema,
        # This is already included via the image, but listed again
        # here for clarity.
        AwsProviderImage.inheritable_aws_image_schema,
        provider_schema.ssh_label,
        aws_label_schema,
    )
    schema = assemble(
        BaseProviderLabel.schema,
        AwsProviderImage.inheritable_aws_image_schema,
        provider_schema.ssh_label,
        aws_label_schema,
    )

    image_flavor_inheritable_schema = assemble(
        AwsProviderImage.inheritable_aws_image_schema,
    )


class AwsProviderSchema(BaseProviderSchema):
    def getLabelSchema(self):
        return AwsProviderLabel.schema

    def getImageSchema(self):
        return AwsProviderImage.schema

    def getFlavorSchema(self):
        return AwsProviderFlavor.schema

    def getInheritableLabelSchema(self):
        return AwsProviderLabel.inheritable_schema

    def getInheritableImageSchema(self):
        return AwsProviderImage.inheritable_schema

    def getInheritableZuulImageSchema(self):
        return AwsProviderImage.inheritable_zuul_schema

    def getInheritableCloudImageSchema(self):
        return AwsProviderImage.inheritable_cloud_schema

    def getInheritableFlavorSchema(self):
        return AwsProviderFlavor.inheritable_schema

    def getProviderSchema(self):
        schema = super().getProviderSchema()
        object_storage = {
            Required(
                'bucket-name',
                doc="""\
                The name of the S3 bucket that should be used when
                importing Zuul images.
                """,
            ): str,
        }

        resource_limits = {k: int for k in ALL_QUOTA_CODES}
        resource_limits[Optional(
            'instances',
            default=vs.UNDEFINED,
            doc="""The number of instances.""",
        )] = int
        resource_limits[Optional(
            'cores',
            default=vs.UNDEFINED,
            doc="""The number of cores.""",
        )] = int
        resource_limits[Optional(
            'ram',
            default=vs.UNDEFINED,
            doc="""The amount of ram, in MiB.""",
        )] = int

        aws_provider_schema = vs.Schema({
            Required(
                'region',
                doc=f"""\
                The name of the `AWS region`_ to interact with.

                .. _`AWS region`: {URL_AWS_REGION}
                """,
            ): str,
            Optional(
                'object-storage',
                doc="""\
                Configuration options related to object storage used
                for image management.
                """,
            ): Nullable(object_storage),
            Optional(
                'resource-limits',
                doc="""\
                Resource limits for this provider.  Configure these
                values to cause Zuul to attempt to limit the resource
                usage.  This can be used to limit Zuul's usage to a
                level below the cloud quota.

                In addition to the options listed below, it is
                possible to configure a limit for any of the quota
                codes supported by AWS.  """,
                default=dict(),
            ): resource_limits,
        })

        return assemble(
            schema,
            aws_provider_schema,
            doc="""\
            The attributes available for configuring an AWS provider
            are below.
            """,
        )


class AwsProvider(BaseProvider, subclass_id='aws'):
    log = logging.getLogger("zuul.AwsProvider")
    schema = AwsProviderSchema().getProviderSchema()

    @property
    def endpoint(self):
        ep = getattr(self, '_endpoint', None)
        if ep:
            return ep
        self._set(_endpoint=self.getEndpoint())
        return self._endpoint

    def parseImage(self, image_config, provider_config, connection):
        return AwsProviderImage(image_config, provider_config)

    def parseFlavor(self, flavor_config, provider_config, connection):
        return AwsProviderFlavor(flavor_config, provider_config)

    def parseLabel(self, label_config, provider_config, connection):
        return AwsProviderLabel(label_config, provider_config)

    def getEndpoint(self):
        return self.driver.getEndpoint(self)

    def validateConfig(self, config):
        for label in config['labels'].values():
            flavor = config['flavors'][label.flavor]
            if flavor.dedicated_host:
                if flavor.market_type == 'spot':
                    raise Exception(
                        "Spot instances can not be used on dedicated hosts")
                if not label.az:
                    raise Exception(
                        "Availability-zone is required for dedicated hosts")

    def getCreateStateMachine(self, node, image_external_id, log):
        # TODO: decide on a method of producing a hostname
        # that is max 15 chars.
        hostname = f"np{node.uuid[:13]}"
        label = self.labels[node.label]
        flavor = self.flavors[label.flavor]
        image = self.images[label.image]
        return AwsCreateStateMachine(
            self,
            self.endpoint,
            node,
            hostname,
            label,
            flavor,
            image,
            image_external_id,
            node.tags,
            log)

    def getDeleteStateMachine(self, node, log):
        return AwsDeleteStateMachine(self.endpoint, node, log)

    def getSnapshotStateMachine(self, node, log):
        return AwsSnapshotStateMachine(self.endpoint, node, log)

    def getEndpointLimits(self):
        # Get the instance and volume types that this provider handles
        limits = self.endpoint.quota_cache.getLimits()
        if limits is None:
            limits = {}
        else:
            limits = limits.quota
        instance_types = {}
        host_types = set()
        volume_types = set()
        for flavor in self.flavors.values():
            if flavor.dedicated_host:
                host_types.add(flavor.instance_type)
            else:
                flavor_instance_types = []
                if flavor.instance_type:
                    flavor_instance_types.append(flavor.instance_type)
                elif flavor.fleet and flavor.fleet.get('instance-types'):
                    # Include instance-types from fleet config if available
                    flavor_instance_types.extend(
                        flavor.fleet['instance-types'])
                for flavor_instance_type in flavor_instance_types:
                    if flavor_instance_type not in instance_types:
                        instance_types[flavor_instance_type] = set()
                    instance_types[flavor_instance_type].add(
                        SPOT if flavor.market_type == 'spot' else ON_DEMAND)
            if flavor.volume_type:
                volume_types.add(flavor.volume_type)
        args = dict(default=math.inf)
        for instance_type in instance_types:
            for market_type_option in instance_types[instance_type]:
                code = self.endpoint._getQuotaCodeForInstanceType(
                    instance_type, market_type_option)
                if code in args:
                    continue
                if not code:
                    continue
                if code not in limits:
                    self.log.warning(
                        "AWS quota code %s for instance type: %s not known",
                        code, instance_type)
                    continue
                args[code] = limits[code]
        for host_type in host_types:
            code = self.endpoint._getQuotaCodeForHostType(host_type)
            if code in args:
                continue
            if not code:
                continue
            if code not in limits:
                self.log.warning(
                    "AWS quota code %s for host type: %s not known",
                    code, host_type)
                continue
            args[code] = limits[code]
        for volume_type in volume_types:
            vquota_codes = VOLUME_QUOTA_CODES.get(volume_type)
            if not vquota_codes:
                self.log.warning(
                    "Unknown quota code for volume type: %s",
                    volume_type)
                continue
            for resource, code in vquota_codes.items():
                if code in args:
                    continue
                if code not in limits:
                    self.log.warning(
                        "AWS quota code %s for volume type: %s not known",
                        code, volume_type)
                    continue
                value = limits[code]
                # Unit mismatch: storage limit is in TB, but usage
                # is in GB.  Translate the limit to GB.
                if resource == 'storage':
                    value *= 1000
                args[code] = value

        return QuotaInformation(**args)

    def getQuotaForLabel(self, label):
        flavor = self.flavors[label.flavor]
        return self.endpoint.getQuotaForLabel(label, flavor)

    def refreshQuotaForLabel(self, label, update):
        flavor = self.flavors[label.flavor]
        return self.endpoint.refreshQuotaForLabel(label, flavor, update)

    def downloadUrl(self, url, path):
        return self.endpoint.downloadUrl(url, path)

    def getImageImportJob(self, url, provider_image, image_name,
                          image_format, metadata, md5, sha256):
        return self.endpoint.getImageImportJob(
            url, provider_image, image_name,
            image_format, metadata, md5, sha256)

    def getImageCopyJob(self, source_provider, provider_image, image_name,
                        image_format, metadata, md5, sha256):
        return self.endpoint.getImageCopyJob(
            source_provider, provider_image, image_name,
            image_format, metadata, md5, sha256)

    def getImageUploadJob(self, provider_image, image_name,
                          image_format, metadata, md5, sha256):
        # TODO this needs to move to the section or connection config
        # since it's used by endpoints.
        bucket_name = self.object_storage.get('bucket_name')
        return self.endpoint.getImageUploadJob(
            provider_image, image_name,
            image_format, metadata, md5, sha256,
            bucket_name)

    def deleteImage(self, external_id):
        self.endpoint.deleteImage(external_id)



# Copyright (C) 2020 Red Hat, Inc
# Copyright (C) 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re

import re2


def filter_allowed_disallowed(
        subjects, allowed_patterns, disallowed_patterns):
    """Filter a list using allowed and disallowed patterns.

    :param list subjects: A list of strings to filter.
    :param allowed_patterns: A list of re2-compatible patterns to allow.
       If empty, all subjects are allowed (see next).
    :param disallowed_patterns: A list of re2-compatible patterns to
       reject.  A more-specific pattern here may override a less-specific
       allowed pattern.  If empty, all allowed subjects will pass.
    """
    ret = []
    for subject in subjects:
        allowed = True
        if allowed_patterns:
            allowed = False
            for pattern in allowed_patterns:
                if re2.match(pattern, subject):
                    allowed = True
                    break
        if allowed and disallowed_patterns:
            for pattern in disallowed_patterns:
                if re2.match(pattern, subject):
                    allowed = False
                    break
        if allowed:
            ret.append(subject)
    return ret


class ZuulRegex:
    def __init__(self, pattern, negate=False):
        self.pattern = pattern
        self.negate = negate
        self.re2_failure = False
        self.re2_failure_message = None
        try:
            o = re2.Options()
            o.log_errors = False
            self.re = re2.compile(pattern, options=o)
        except re2.error as e:
            # Compile under re first to find out if this is also a
            # PCRE error, which should take precedence.
            self.re = re.compile(pattern)
            # If it compiled okay, then the problem is re2 vs pcre
            self.re2_failure = True
            if e.args and len(e.args) == 1:
                if isinstance(e.args[0], bytes):
                    self.re2_failure_message = e.args[0].decode('utf8')
                elif isinstance(e.args[0], str):
                    self.re2_failure_message = e.args[0]

    def __eq__(self, other):
        return (isinstance(other, ZuulRegex) and
                self.pattern == other.pattern and
                self.negate == other.negate)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(json.dumps(self.toDict(), sort_keys=True))

    def match(self, subject):
        if self.negate:
            return not self.re.match(subject)
        return self.re.match(subject)

    def fullmatch(self, subject):
        if self.negate:
            return not self.re.fullmatch(subject)
        return self.re.fullmatch(subject)

    def search(self, subject):
        if self.negate:
            return not self.re.search(subject)
        return self.re.search(subject)

    def toDict(self):
        # This is used in user-facing serialization, like zuul-web, to
        # match job syntax.
        return {
            "regex": self.pattern,
            "negate": self.negate,
        }

    def serialize(self):
        return {
            "pattern": self.pattern,
            "negate": self.negate,
        }

    @classmethod
    def deserialize(cls, data):
        o = cls(data['pattern'], data['negate'])
        return o



# Copyright 2013 Rackspace Australia
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import voluptuous as v

from zuul.lib.logutil import get_annotated_logger
from zuul.reporter import BaseReporter


class SMTPReporter(BaseReporter):
    """Sends off reports to emails via SMTP."""

    name = 'smtp'
    log = logging.getLogger("zuul.SMTPReporter")

    def report(self, item, phase1=True, phase2=True):
        """Send the compiled report message via smtp."""
        if not phase1:
            return
        log = get_annotated_logger(self.log, item.event)
        message = self._formatItemReport(item)

        log.debug("Report %s, params %s, message: %s",
                  item, self.config, message)

        from_email = self.config['from'] \
            if 'from' in self.config else None
        to_email = self.config['to'] \
            if 'to' in self.config else None

        if 'subject' in self.config:
            subject = self.config['subject'].format(
                change=item.changes[0],
                changes=item.changes,
                pipeline=item.manager.pipeline.getSafeAttributes())
        else:
            subject = "Report for changes {changes} against {ref}".format(
                changes=' '.join([str(c) for c in item.changes]),
                ref=' '.join([c.ref for c in item.changes]))

        self.connection.sendMail(subject, message, from_email, to_email,
                                 zuul_event_id=item.event)
        return []


def getSchema():
    smtp_reporter = v.Schema({
        'to': str,
        'from': str,
        'subject': str,
    })
    return smtp_reporter



# Copyright 2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Print out a representation of a provider configuration schema.

import argparse
import textwrap

from zuul.driver.openstack import openstackprovider
from zuul.driver.aws import awsprovider
from zuul.driver.azure import azureprovider
from zuul.driver.static import staticprovider
from zuul.lib.voluputil import AsList, Nullable, Constant, UNDEFINED

import voluptuous as vs
import yaml


class IndentedListDumper(yaml.Dumper):
    def increase_indent(self, flow=False, *args, **kwargs):
        return super().increase_indent(flow=flow, indentless=False)


class SchemaWalker:
    def __init__(self, schema, driver, doc=None):
        self.schema = schema
        self.driver = driver
        self.doc = doc or getattr(schema, 'doc', None)

    def toDict(self):
        if isinstance(self.schema, str):
            return dict(type="const", value=self.schema, doc=self.doc)
        if self.schema is True:
            return dict(type="const", value=self.schema, doc=self.doc)
        if self.schema is False:
            return dict(type="const", value=self.schema, doc=self.doc)
        if self.schema == str:  # noqa
            return dict(type="str", doc=self.doc)
        if self.schema == int:  # noqa
            return dict(type="int", doc=self.doc)
        if self.schema == bool:  # noqa
            return dict(type="bool", doc=self.doc)
        if self.schema == float:  # noqa
            return dict(type="float", doc=self.doc)
        if self.schema == None:  # noqa
            return dict(type="const", value='null', doc=self.doc)
        if isinstance(self.schema, Constant):
            w = SchemaWalker(self.schema.schema, self.driver)
            return w.toDict()
        if isinstance(self.schema, Nullable):
            w = SchemaWalker(self.schema.schema, self.driver, self.doc)
            return w.toDict()
        if isinstance(self.schema, vs.Schema):
            w = SchemaWalker(self.schema.schema, self.driver, self.doc)
            return w.toDict()
        if isinstance(self.schema, AsList):
            # Use the first alt (the list); ignore the second.
            w = SchemaWalker(self.schema.schema.validators[0],
                             self.driver, self.doc)
            return w.toDict()
        if isinstance(self.schema, vs.Any):
            ret = []
            for v in self.schema.validators:
                w = SchemaWalker(v, self.driver)
                v = w.toDict()
                ret.append(v)
            islist = False
            if len(ret) == 1 and isinstance(ret[0], list):
                ret = ret[0]
                islist = True
            return dict(alts=ret, doc=self.doc, list=islist)
        if isinstance(self.schema, vs.All):
            schemas = []
            for v in self.schema.validators:
                w = SchemaWalker(v, self.driver)
                v = w.toDict()
                schemas.append(v)
            ret = {}
            for x in schemas:
                if isinstance(x, dict):
                    ret.update(x)
            if not ret.get('doc'):
                ret['doc'] = self.doc
            return ret
        if isinstance(self.schema, vs.Union):
            ret = []
            for v in self.schema.validators:
                w = SchemaWalker(v, self.driver)
                v = w.toDict()
                ret.append(v)
            return dict(alts=ret, doc=self.doc)
        if isinstance(self.schema, dict):
            ret = {}
            for k, v in self.schema.items():
                name = str(k)
                if '_' in name:
                    continue
                w = SchemaWalker(v, self.driver, getattr(k, 'doc', None))
                v = w.toDict()
                if (hasattr(k, 'default') and
                    k.default not in [UNDEFINED, vs.UNDEFINED]):
                    d = k.default()
                    if (d not in [UNDEFINED, vs.UNDEFINED] and
                        d != dict() and
                        d != list()):
                        v['default'] = d
                ret[name] = v
            return dict(type='dict', doc=self.doc, value=ret)
        if isinstance(self.schema, list):
            elements = []
            for v in self.schema:
                w = SchemaWalker(v, self.driver)
                v = w.toDict()
                if isinstance(v, list):
                    elements.extend(v)
                else:
                    elements.append(v)
            if len(elements) == 1 and not elements[0].get('doc'):
                ret = elements[0].copy()
                ret['doc'] = self.doc
                ret['list'] = True
                return ret
            return elements

    def _attr(self, indent, name, data):
        indent1 = indent * '   '
        indent2 = (indent + 1) * '   '
        kind = data.get('type')
        if not kind and data.get('list'):
            kind = 'list'
        out = f"{indent1}.. attr:: {name}\n"
        if kind and kind not in ['const']:
            out += f"{indent2}:type: {kind}\n"
        if 'default' in data:
            out += f"{indent2}:default: {data['default']}\n"
        if data.get('required'):
            out += f"{indent2}:required:\n"
        out += '\n'
        if data.get('doc'):
            doc = textwrap.dedent(data['doc'])
            doc = textwrap.indent(doc, indent2)
            doc = doc.replace("$driver", self.driver)
            out += doc
            while not out.endswith('\n\n'):
                out += '\n'
        else:
            out += "TODO: document\n\n"
        if data.get('value'):
            out += self._toDoc(indent + 1, data)
        if data.get('alts'):
            for alt in data['alts']:
                # This is not generic, it only works for images
                if name == 'images':
                    disc = alt['value']['type']['alts'][0]['value']
                    altname = f"{name}[{disc}]"
                    out += self._attr(indent, altname, alt)
                else:
                    out += self._toDoc(indent + 1, alt)
        return out

    def _value(self, indent, data):
        indent1 = indent * '   '
        indent2 = (indent + 1) * '   '
        value = data['value']
        out = f"{indent1}.. value:: {value}\n"
        out += '\n'
        if data.get('doc'):
            doc = textwrap.dedent(data['doc'])
            doc = textwrap.indent(doc, indent2)
            doc = doc.replace("$driver", self.driver)
            out += doc
            while not out.endswith('\n\n'):
                out += '\n'
        else:
            out += "TODO: document\n\n"
        return out

    def _toDoc(self, indent, data):
        out = ''
        if data.get('type') == 'dict':
            value = data['value']
            for k in sorted(value.keys()):
                v = value[k]
                # Free-form dicts like "tags" end up with this
                if k == "<class 'str'>":
                    continue
                # Special case for AWS quota resources
                if k.startswith("L-"):
                    continue
                if v.get('doc') == "UNDOCUMENTED":
                    continue
                out += self._attr(indent, k, v)
        elif data.get('type') == 'const':
            out += self._value(indent, data)
        return out

    def toDoc(self):
        data = self.toDict()
        out = ''

        provider = {
            'value': {
                f'provider[{self.driver}]': data,
            },
            'type': 'dict',
        }
        out += self._toDoc(0, provider)
        return out


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="Print provider configuration schema")
    parser.add_argument('--yaml', dest='yaml', action='store_true',
                        help='Output YAML')
    parser.add_argument(dest='driver',
                        help='The driver to inspect')
    args = parser.parse_args()
    if args.driver == 'aws':
        ps = awsprovider.AwsProviderSchema()
    elif args.driver == 'openstack':
        ps = openstackprovider.OpenstackProviderSchema()
    elif args.driver == 'static':
        ps = staticprovider.StaticProviderSchema()
    elif args.driver == 'azure':
        ps = azureprovider.AzureProviderSchema()
    else:
        raise Exception("Unknown driver")
    s = ps.getProviderSchema()
    w = SchemaWalker(s, args.driver)
    if args.yaml:
        out = w.toDict()
        print(yaml.dump(out, Dumper=IndentedListDumper,
                        default_flow_style=False))
    else:
        out = w.toDoc()
        print(out)
